// qlcp-panel.js
'use strict';

//****named constants****************************************************************************************************************************************************************
const SW_VERSION = "4.13.0";
const CPF_VERSION = 5;

const SERVER_PORT_DEFAULT = 9999;                         // Default port for QLCP-NET server
const NUM_INDICATION_BITS_DEFAULT = 512;                  // Default number of indication bits

const SERVER_PORT_PARAM = "serverPort";                   // Parameter key for QLCP-NET server port
const NUM_INDICATION_BITS_PARAM = "numIndicationBits";    // Parameter key for number of indication bits
const AUTO_OPEN_PARAM = "openfile";                       // Parameter key for panel file auto open flag
const READ_ONLY_PARAM = "readOnly";                       // Parameter key for read only (disable panel edit/save) flag

const OPACITY_KNOWN = 1.0;                                // Component opacity level for 'known' state
const OPACITY_UNKNOWN = 0.3;                              // Component opacity level for 'unknown' (or 'uncertain') state

//****global variables***************************************************************************************************************************************************************
let QnetClientThread;                   // Web worker object for qnet-client.js thread that sends controls to and requests indications from the server.
let linkOk = false;                     // Link health between client / server. Needed to detect transitions in link state determined by the web worker (qnet-client.js).
let redrawComponents = false;           // Indicates component layer needs to be drawn/updated (occurs at a browser determined interval in requestAnimationFrame processing.)
let developerMode = false;              // True when connected to the QLCP-NET Developer Tool server via localhost.
let designMode = false;                 // True when in design mode (able to edit panel layout and bit assignments.)
let controlOptionsStr = "";             // Control label/bit name string used to set the innerHTML of properties select options.
let indicationOptionsStr = "";          // Indication label/bit name string used to set the innerHTML of properties select options.
let nbrRows = 60;                       // Max rows in editor grid.
let nbrCols = 182;                      // Max columns in editor grid.
let numIndicationBits = NUM_INDICATION_BITS_DEFAULT;  // Number of indication bits
let controlBits = new Array(512);       // Stores control bit states for latched controls (e.g. toggle switch, keyswitch, rotary switch ,etc.
let requestFileName = false;            // Name of saved panel file to open; requested from the developer tool (server)
let hostUrl = window.location.host;     // IP address of the QLCP-NET server that is entered for the URL in the browser
let serverPort = SERVER_PORT_DEFAULT;   // IP port to use when accessing the QLCP-NET server
let passcode = "";                      // Passcode for locked controls.
let lastClickTime = new Date();         // Time when user last clicked on a control component (for passcode functionality.)
let initLatchedSwitches = false;        // Set true when changing between run and design mode to initialize latched switches
let selectedId = "";                    // Id of currently selected object (built from idNbr and prefix)
let selectedComponents = new Array();   // Contains the component objects captued in a selection.
let compInSelectedGroup = false;        // True when a given component is in a selected group
let mouseDownOnComponent = false;       // True if there was a mouse down event on any component
let initX = 0;                          // Initial x coordinate of mouse position
let initY = 0;                          // Initial y coordinate of mouse position
let draggingToSelectComponents = false; // True while dragging mouse to select a group of components
let selectionBoxCreated = false;        // True if the rectangle shape for the selection box is added to the component layer
let selectionBox = null;                // Selection box (rectangle shape) object displayed while making a group selection
let panelCRC = 0;
let allow_bgImg_selection = false;
let lastTextAlign = 'left';
let autoOpen = false;                   // True when request is made to immediately request a panel file when client is launched
let readOnly = false;                   // True when the ability to edit a panel and save a panel is disabled

let ctrlKeyDown = false;                // Capture when control key is pressed
let idNbr = 0;                          // The next id number to assign a component.

const defaultFillColor = 'rgba(85,85,85,1)';

// Key:Value pairs that track drop modes for each component type
// The boolean value is true when the drop mode for that component type is active.
let dropMode = {
  text: false,
  LED: false,
  pb: false,
  pb_square: false,
  togSw: false,
  togSwMom: false,
  quadrilateral: false,
  arrow: false,
  timer: false,
  keyswitch: false,
  rotaryswitch: false,
  none: true
};

// Key:Value pairs that map button ids to component types.
const buttonToComponent = {
  text_button: "text",
  led_button: "LED",
  pb_button: "pb",
  pb_square_button: "pb_square",
  togSw_button: "togSw",
  quadrilateral_button: "quadrilateral",
  arrow_button: "arrow",
  timer_button: "timer",
  keyswitch_button: "keyswitch",
  rotaryswitch_button: "rotaryswitch"
};

// Used for display date and time in the LCP header div
const monthNames = ["JAN", "FEB", "MAR", "APR", "MAY", "JUN", "JUL", "AUG", "SEP", "OCT", "NOV", "DEC"];

// Colors used for drawing components and text, and assigning indications.
const stdColors = ["red", "yellow", "lime",
                  "green", "orange", "blue",
                  "white", "black", "cyan",
                  "dodgerblue", "gray", "dimgray",
                  "transparent"];
const colorVals =  ["rgba(255,0,0,1)", "rgba(255,255,0,1)", "rgba(0,255,0,1)",
                  "rgba(0,128,0,1)", "rgba(255,165,0,1)", "rgba(0,0,255,1)",
                  "rgba(255,255,255,1)", "rgba(0,0,0,1)", "rgba(0,255,255,1)",
                  "rgba(30,144,255,1)", "rgba(128,128,128,1)", "rgba(105,105,105,1)",
                  "rgba(0,0,0,0)"];

function RGBAToHexA(rgba) {
  let ex = /^rgba\((((((((1?[1-9]?\d)|10\d|(2[0-4]\d)|25[0-5]),\s?)){3})|(((([1-9]?\d(\.\d+)?)|100|(\.\d+))%,\s?){3}))|(((((1?[1-9]?\d)|10\d|(2[0-4]\d)|25[0-5])\s){3})|(((([1-9]?\d(\.\d+)?)|100|(\.\d+))%\s){3}))\/\s)((0?\.\d+)|[01]|(([1-9]?\d(\.\d+)?)|100|(\.\d+))%)\)$/i;
  if (ex.test(rgba)) {
    let sep = rgba.indexOf(",") > -1 ? "," : " ";
    rgba = rgba.substr(5).split(")")[0].split(sep);

    // strip the slash if using space-separated syntax
    if (rgba.indexOf("/") > -1)
      rgba.splice(3,1);

    for (let R in rgba) {
      let r = rgba[R];
      if (r.indexOf("%") > -1) {
        let p = r.substr(0,r.length - 1) / 100;

        if (R < 3) {
          rgba[R] = Math.round(p * 255);
        } else {
          rgba[R] = p;
        }
      }
    }

    let r = (+rgba[0]).toString(16),
      g = (+rgba[1]).toString(16),
      b = (+rgba[2]).toString(16),
      a = Math.round(+rgba[3] * 255).toString(16);

    if (r.length == 1)
      r = "0" + r;
    if (g.length == 1)
      g = "0" + g;
    if (b.length == 1)
      b = "0" + b;
    if (a.length == 1)
      a = "0" + a;

    return "#" + r + g + b + a;

  } else {
    return "Invalid input color";
  }
} // end RGBAToHex()

function hexAToRGBA(h,isPct) {
  let ex = /^#([\da-f]{4}){1,2}$/i;
  if (ex.test(h)) {
    let r = 0, g = 0, b = 0, a = 1;
    isPct = isPct === true;

    if (h.length == 5) {
      r = "0x" + h[1] + h[1];
      g = "0x" + h[2] + h[2];
      b = "0x" + h[3] + h[3];
      a = "0x" + h[4] + h[4];

    } else if (h.length == 9) {
      r = "0x" + h[1] + h[2];
      g = "0x" + h[3] + h[4];
      b = "0x" + h[5] + h[6];
      a = "0x" + h[7] + h[8];
    }
    a = +(a / 255).toFixed(3);
    if (isPct) {
      r = +(r / 255 * 100).toFixed(1);
      g = +(g / 255 * 100).toFixed(1);
      b = +(b / 255 * 100).toFixed(1);
      a = +(a * 100).toFixed(1);
    }

    return "rgba("+ (isPct ? r + "%," + g + "%," + b + "%" + "," + a : +r + "," + +g + "," + +b + "," + a) + ")";

  } else {
    return "Invalid input color";
  }
} // end hexAToRGBA()

/******************************************************************************/
/******************************************************************************/
//Create the stage for component and text.
//44 pixels is based on the grid size, overlap of grids, etc.
let stage = new Konva.Stage({
  container: 'component_container',
  width: nbrCols * 44,
  height: nbrRows * 44
});

//create the component layer and add it to the stage
let component_layer = new Konva.Layer();
let bgImage_layer = new Konva.Layer();
stage.add(bgImage_layer, component_layer);

let scrollContainer = document.getElementById('scroll_container');
scrollContainer.addEventListener('scroll', function () {
  let dx = scrollContainer.scrollLeft;
  let dy = scrollContainer.scrollTop;
  stage.container().style.transform = 'translate(' + dx + 'px, ' + dy + 'px)';
  stage.x(-dx);
  stage.y(-dy);
  stage.batchDraw();
});

//The following three mouse event routines are for selecting multiple components on the stage
//(on the component layer). The reason the event listeners are bound to the stage is so
//they only fire when the event occurs in the stage area, not the title, menu, or footer areas.

//This event depends on bubbling of events from the component elements on the component layer.
//It depends on the component setting the selected Id and mouseDownOnComponent flag first.
//This routine is only used for design mode.
stage.getContent().addEventListener("mousedown", function () {
  if (designMode) {
    //If the mouse down event was on the stage and not on a component, and if
    //there is a drop mode active, add the component to the stage.
    //AddComponent function will check drop mode to determine which component to add.
    if (!mouseDownOnComponent)  {
      allow_bgImg_selection = true;
      // console.log("clear selected group of components");
      //Need to clear any previously selected group of components
      ClearSelectedComponents();

      if (dropMode['none']) {

        draggingToSelectComponents = true;
        // stop selection box when dragging background image.
        if (bgImage_layer.hasChildren()) {
          if (bgImage_layer.getChildren()[0].draggable()) {
            draggingToSelectComponents = false;
          }
        }
      }

      //get the current mouse position
      let mousepos = stage.getPointerPosition();

      //If drop mode is not none, add the component to the layer at the current mouse position.
      if (!dropMode['none']) {
        let Component = AddComponent(Math.round(mousepos.x), Math.round(mousepos.y), ++idNbr);
        ShowSelectedComponent(Component);
      } else {
        ClearPreviousSelection();
      }
    } else {
      //mouse down was on a component
      allow_bgImg_selection = false;
      //If multiple components were selected (array is not empty),
      //set flag if a component within the group has been selected.
      for (let i = 0; i < selectedComponents.length; i++) {
        if (selectedComponents[i].attrs.id === selectedId) {
          compInSelectedGroup = true;
        }
      }
      //If the mouse down event hit a component in a selected group,
      //get the mouse position to start keeping track of the move distance (delta x, delta y).
      if (compInSelectedGroup) {
        let mousepos = stage.getPointerPosition();
        initX = Math.round(mousepos.x);
        initY = Math.round(mousepos.y);
      } else {
        //Need to clear any previously selected group of components
        ClearSelectedComponents();
      }
    }
  }
}); // end stage mousedown

//This function depends on requestAnimationFrame to redraw the component layer.
//This routine is only used for design mode.
stage.getContent().addEventListener('mousemove', function () {
  if (designMode) {
    if (draggingToSelectComponents) {
      CreateSelectionBox();
    }

    //If a group is selected and a component in the group has been selected,
    //start moving the group of selected components.
    if (selectedComponents.length > 0 && compInSelectedGroup) {
      if (TogglePositionLock.positionLocked != true) {

        //Get the current mouse position and determine the delta x and y values.
        let mousepos = stage.getPointerPosition();
        let deltaX = Math.round(mousepos.x) - initX;
        let deltaY = Math.round(mousepos.y) - initY;

        //Move each of the components in the selection.
        for (let i = 0; i < selectedComponents.length; i++) {
          //Move all other components. The selected one will move when dragged.
          if (selectedComponents[i].attrs.id !== selectedId) {
            selectedComponents[i].position({
              x: selectedComponents[i].getPosition().x + deltaX,
              y: selectedComponents[i].getPosition().y + deltaY
            });
          }
        }

        //Reset coordinates for the next move event.
        initX = Math.round(mousepos.x);
        initY = Math.round(mousepos.y);
      }
    }

    //show the currently selected component's x,y position
    if (selectedComponents.length == 0 && !mouseDownOnComponent) {
      if (bgImage_layer.hasChildren()) {
        if (bgImage_layer.getChildren()[0].draggable()) {
          document.getElementById("position_display_textbox").value = "  x = " +
          Math.round(bgImage_layer.getChildren()[0].x()) + ", y = " + Math.round(bgImage_layer.getChildren()[0].y());
        }
      }
    }

    if (selectedId !== "") {
      let component = stage.find('#'+ selectedId)[0];
      ShowComponentPosition(component);

      if ($( "#properties_dialog" ).dialog( "isOpen" )) {
        document.getElementById("x-pos-value").value = component.attrs.x;
        document.getElementById("y-pos-value").value = component.attrs.y;
        document.getElementById('z-value').value = component.getZIndex()+1;
      }
    }
  }
}); // end stage mousemove

//This function depends on requestAnimationFrame to redraw the component layer.
//This routine is only used for design mode.
stage.getContent().addEventListener("mouseup", function ()
{
  if (designMode)
  {
    //When the mouse is released after dragging to select
    //components (flag is still true), determine which
    //components are within the selection box and show them as selected.
    if (draggingToSelectComponents && !mouseDownOnComponent)
    {
      draggingToSelectComponents = false;

      //clear the selection box
      if (selectionBoxCreated)
      {
        ShowSelectedGroup();
        selectionBox.destroy();
        selectionBoxCreated = false;
      }
    }

    //If there is a group of selected components...
    if (selectedComponents.length > 0)
    {
      //clear the selected Id and previously selected Id.
      selectedId = "";
    }

    //clear flags for next selection
    compInSelectedGroup = false;
    mouseDownOnComponent = false;
  }
}); // end stage mouseup

// For now use double click to enable/disable background image dragging.
stage.getContent().addEventListener("dblclick", () => {
  if (designMode) {
    if (allow_bgImg_selection)
    {
      if (bgImage_layer.hasChildren()) {

        draggingToSelectComponents = false;

        let bgImg = bgImage_layer.getChildren()[0];

        if (bgImg.draggable()) {
          bgImg.draggable(false);
          bgImg.strokeEnabled(false);
        } else {
          bgImg.draggable(true);
          bgImg.strokeEnabled(true);
          document.getElementById("position_display_textbox").value = "  x = " +
          Math.round(bgImage_layer.getChildren()[0].x()) + ", y = " +
          Math.round(bgImage_layer.getChildren()[0].y());
        }
        bgImage_layer.draw();
      }
    }
  }
}); // end stage doubleclick

/******************************************************************************/
/******************************************************************************/

function UpdateControlLabels(data) {
  let controlLabels = data.split(",");
  controlOptionsStr = "<option>None</option>";

  //color-code groups of status names (128 per) to make it easier for multiple-server soft panel designs
  let listBackgroundColor = "rgba(198,224,180,1)";
  for(let i=0; i < controlLabels.length - 1; i++) {
    if (i === 128) {
      listBackgroundColor = "rgba(255,230,153,1)";
    }
    if (i === 256) {
      listBackgroundColor = "rgba(189,215,238,1)";
    }
    if (i === 384) {
      listBackgroundColor = "rgba(248,203,173,1)";
    }

    controlOptionsStr += "<option style='background-color: " + listBackgroundColor + "'>" + controlLabels[i] + "</option>";
    controlBits[i] = 0; //initialize control bit state
  }
} // end UpdateControlLabels()

function UpdateIndicationLabels(data)
{
  let indicationLabels = data.split(",");
  indicationOptionsStr = "<option>None</option>";
  let listBackgroundColor = "rgba(198,224,180,1)";

  //color-code groups of status names (128 per) to make it easier for multiple-server soft panel designs
  for(let i=0; i < indicationLabels.length - 1; i++)
  {
    if (i === 128) {
      listBackgroundColor = "rgba(255,230,153,1)";
    }
    if (i === 256) {
      listBackgroundColor = "rgba(189,215,238,1)";
    }
    if (i === 384) {
      listBackgroundColor = "rgba(248,203,173,1)";
    }
    indicationOptionsStr = indicationOptionsStr + "<option style='background-color: " + listBackgroundColor + "'>" + indicationLabels[i] + "</option>";
  }
} // end UpdateIndicationLabels()

//This function is periodically called using requestAnimationFrame to draw the component
//layer. This is required when changes are made to the appearance of a component.
//Since design mode will typically be used on a fast enough laptop PC, the component layer
//is drawn as frequently as the browser can. For run mode, the redrawComponents flag is used
//to only draw the layer when necessary.
function UpdateComponentLayer() {
  if (designMode)
  {
    component_layer.draw();
  }
  else if (redrawComponents)
  {
    component_layer.draw();
    redrawComponents = false;
  }

  //requestAnimationFrame works using recursion
  window.requestAnimationFrame(UpdateComponentLayer);
} // end UpdateComponentLayer()

function InitGrid() {
  //Create the divs for the grid lines
  let htmlGridLines = "";

  for (let rowIndex=0; rowIndex < nbrRows; rowIndex++)
  {
    for(let colIndex=0; colIndex < nbrCols; colIndex++)
    {
      htmlGridLines += '<div class="gridLines" id="s' + rowIndex + ',' + colIndex + '" style="float: left"></div>';
    }
    // start the next row (no id assigned to cell, just a fixed block got to next row)
    htmlGridLines += '<div class="gridLines"></div>';
  }

  document.getElementById("grid_lines").innerHTML = htmlGridLines;

  if (designMode === false)
  {
    //Hide grid borders.
    document.getElementById("grid_lines").style.visibility = "hidden";
  }
} // end InitGrid()

function InitDialogBoxes() {
  $( "#properties_dialog" ).dialog({ autoOpen: false });
  $( "#properties_dialog" ).dialog({ minWidth: "510" });
  $( "#properties_dialog" ).dialog({ height: "700" });
  $( "#properties_dialog" ).dialog({ resizable: true });
  $( "#properties_dialog" ).dialog({ appendTo: "#buttonBackground" });
  $( "#properties_dialog" ).dialog({ close: ClearPreviousSelection });

  $( "#controls_dialog" ).dialog({ autoOpen: false });
  $( "#controls_dialog" ).dialog({ width: "auto" });
  $( "#controls_dialog" ).dialog({ height: "auto" });
  $( "#controls_dialog" ).dialog({ resizable: false });
  $( "#controls_dialog" ).dialog({ appendTo: "#headerWrapper" });

  $( "#clear_all_dialog" ).dialog({ autoOpen: false });
  $( "#clear_all_dialog" ).dialog({ width: "auto" });
  $( "#clear_all_dialog" ).dialog({ height: "auto" });
  $( "#clear_all_dialog" ).dialog({ resizable: false });
  $( "#clear_all_dialog" ).dialog({ modal: true });
  $( "#clear_all_dialog" ).dialog({ appendTo: "#buttonBackground" });

  $( "#message_dialog" ).dialog({ autoOpen: false });
  $( "#message_dialog" ).dialog({ width: "auto" });
  $( "#message_dialog" ).dialog({ height: "auto" });
  $( "#message_dialog" ).dialog({ resizable: false });
  $( "#message_dialog" ).dialog({ modal: true });
  $( "#message_dialog" ).dialog({ appendTo: "#headerWrapper" });

  $( "#passcode_dialog" ).dialog({ autoOpen: false });
  $( "#passcode_dialog" ).dialog({ width: "auto" });
  $( "#passcode_dialog" ).dialog({ height: "auto" });
  $( "#passcode_dialog" ).dialog({ resizable: false });
  $( "#passcode_dialog" ).dialog({ modal: true });
  $( "#passcode_dialog" ).dialog({ appendTo: "#headerWrapper" });
} // end InitDialogBoxes()

function CloseAllDialogBoxes() {
  if ($( "#properties_dialog" ).dialog( "isOpen" )) {
    $( "#properties_dialog" ).dialog( "close" );
  }

  if ($( "#controls_dialog" ).dialog( "isOpen" )) {
    $( "#controls_dialog" ).dialog( "close" );
  }

  if ($( "#clear_all_dialog" ).dialog( "isOpen" )) {
    $( "#clear_all_dialog" ).dialog( "close" );
  }

  if ($( "#message_dialog" ).dialog( "isOpen" )) {
    $( "#message_dialog" ).dialog( "close" );
  }

  if ($( "#passcode_dialog" ).dialog( "isOpen" )) {
    $( "#passcode_dialog" ).dialog( "close" );
  }
} // end CloseAllDialogBoxes()

function DesignMode() {
  designMode = true;

  // Hide the CRC - would like to figure out a way to keep
  // visible with live updating CRC as desing happens.
  document.getElementById("panel-crc").style.visibility = "hidden";

  //show button background
  document.getElementById("buttonBackground").style.visibility = "visible";

  //hide run mode buttons
  document.getElementById("fullscreen_button").style.visibility = "hidden";
  document.getElementById("design_button").style.visibility = "hidden";
  document.getElementById("send_all_button").style.visibility = "hidden";
  document.getElementById("clear_controls_button").style.visibility = "hidden";

  //show design mode buttons
  document.getElementById("designButtons").style.visibility = "visible";
  document.getElementById("clear_all_button").style.visibility = "visible";

  SetPanelStyle(false);

  //hide buttons for editing and saving panels if in 'read only' mode
  if (readOnly) {
    HideToolsButtons();
    document.getElementById("save_button").style.visibility = "hidden";
  }

  //Show grid borders
  document.getElementById("grid_lines").style.visibility = "visible";

  //close dialog boxes if they are open
  CloseAllDialogBoxes();

  //turn off all latched switches
  let components = component_layer.getChildren();
  components.draggable(!TogglePositionLock.positionLocked);
  let clickEvent = document.createEvent('MouseEvents');
  clickEvent.initEvent('mousedown', true, true);

  //set flag to tell the component event handlers that the events we are about to fire are for initialization
  initLatchedSwitches = true;

  for (let i = 0; i < components.length; i++) {
    let comp = components[i];

    if (typeof comp.attrs.type !== 'undefined') {
      if ((comp.attrs.type === 'togSw' || comp.attrs.type === 'keyswitch' || comp.attrs.type === 'rotaryswitch') && comp.attrs.state === true) {
        //turn it off by firing a click event to toggle the state
        comp.dispatchEvent(clickEvent);
        comp.draggable(false);
      }
    }
  }

  initLatchedSwitches = false;

  //clear selection of the last latched switch that may have been selected by initialization sequence
  ClearPreviousSelection();

  //change styles to maximize stage and remove scrollable div area
  document.getElementsByTagName("body")[0].style.overflow = "visible";
  document.getElementById("scroll_container").style.top = "93px";
  document.getElementById("scroll_container").style.width = "16020px";
  document.getElementById("scroll_container").style.height = "10020px";
} // end DesignMode()

function RunMode() {
  designMode = false;

  if (bgImage_layer.hasChildren()) {

    let bgImg = bgImage_layer.getChildren()[0];
    bgImg.draggable(false);
    bgImg.strokeEnabled(false);
    bgImage_layer.draw();
  }

  //hide design mode buttons
  document.getElementById("designButtons").style.visibility = "hidden";

  //hide tools buttons
  HideToolsButtons();

  //show run mode buttons
  document.getElementById("fullscreen_button").style.visibility = "visible";
  document.getElementById("design_button").style.visibility = "visible";

  // Hide button menu and run buttons
  document.getElementById("send_all_button").style.visibility = "hidden";
  document.getElementById("clear_controls_button").style.visibility = "hidden";

  //Hide grid borders
  document.getElementById("grid_lines").style.visibility = "hidden";

  //hide any dialog boxes from design mode if they are open
  CloseAllDialogBoxes();

  ClearSelectedComponents();

  //deselect any buttons for dropping components
  DeselectDropComponentButtons();

  // Serialize the panel just to update the CRC based on any changes made since open/save.
  SerializePanelToString();
  document.getElementById("panel-crc").style.visibility = "visible";

  //Initialize control bit states and send all controls...
  //First, clear all states in the controlBits array.
  for (let i=0; i < controlBits.length; i++) {
    controlBits[i] = 0;
  }

  //Now, turn off all toggle switches and set their off states to 1.
  let components = component_layer.getChildren();
  components.draggable(false);

  let clickEvent = document.createEvent('MouseEvents');
  clickEvent.initEvent('mousedown', true, true);

  //set flag to tell the component event handlers that the events we are about to fire are for initialization
  initLatchedSwitches = true;

  for (let i = 0; i < components.length; i++) {
    let comp = components[i];

    //if component should be locked, lock it
    if (typeof comp.attrs.unlocked !== 'undefined') {
      comp.attrs.unlocked = false;
    }

    if (typeof comp.attrs.type !== 'undefined') {
      if (comp.attrs.type === 'togSw') {
        comp.dispatchEvent(clickEvent); //turn switch on
        comp.dispatchEvent(clickEvent); //turn switch off
      }
      if (comp.attrs.type === 'keyswitch' || comp.attrs.type === 'rotaryswitch') {
        comp.dispatchEvent(clickEvent); //set to position 1 (up)
      }
    }
  }

  initLatchedSwitches = false;

  //change styles to minimize stage in a scrollable area
  document.getElementsByTagName("body")[0].style.overflow = "hidden";
  document.getElementById("scroll_container").style.top = "50px";
  document.getElementById("scroll_container").style.width = "calc(100% - 2px)";
  document.getElementById("scroll_container").style.height = "calc(100vh - 82px)";
  document.getElementById("buttonBackground").style.visibility = "hidden";

  SendAllControls();
} // end RunMode()

function SetPanelStyle(userChangedSelection) {
  //set tools up for Components style
  ShowComponentTools();

  //deselect any buttons for dropping components
  DeselectDropComponentButtons();

  //hide any dialog boxes from design mode if they are open
  if ($( "#properties_dialog" ).dialog( "isOpen" )) {
    $( "#properties_dialog" ).dialog( "close" );
  }

} // end SetPanelStyle()

function ShowComponentTools() {
  document.getElementById("text_button").style.visibility = "visible";
  document.getElementById("componentButtons").style.visibility = "visible";
  document.getElementById("increase_button").style.visibility = "visible";
  document.getElementById("decrease_button").style.visibility = "visible";
  document.getElementById("remove_background_button").style.visibility = "visible";
  document.getElementById("properties_button").style.visibility = "visible";
  document.getElementById("clear_all_button").style.visibility = "visible";
  document.getElementById("position_display").style.visibility = "visible";
  document.getElementById("lock_positions_button").style.visibility = "visible";
} // end ShowComponentTools()

function HideToolsButtons() {
  document.getElementById("text_button").style.visibility = "hidden";
  document.getElementById("componentButtons").style.visibility = "hidden";
  document.getElementById("increase_button").style.visibility = "hidden";
  document.getElementById("decrease_button").style.visibility = "hidden";
  document.getElementById("remove_background_button").style.visibility = "hidden";
  document.getElementById("properties_button").style.visibility = "hidden";
  document.getElementById("clear_all_button").style.visibility = "hidden";
  document.getElementById("position_display").style.visibility = "hidden";
  document.getElementById("lock_positions_button").style.visibility = "hidden";
} // HideToolsButtons()

function SendControl(bit, state) {
  controlBits[bit] = state; // State of the controlBits array for this bit value is updated here to allow temp this method to be called with temp data.
  QnetClientThread.postMessage( {
    devMode:developerMode,
    url:hostUrl,
    serverPort:serverPort,
    sendControl:true,
    bitNbr:bit,
    bitState:state,
    sendAllControls:false,
    cntrlBits:controlBits,
    numIndicationBits:numIndicationBits } );
} // end SendControl()

function SendAllControls() {
  // It is up to the caller to make sure the control bits array is updated prior to sending all.
  QnetClientThread.postMessage( {
    devMode:developerMode,
    url:hostUrl,
    serverPort:serverPort,
    sendControl:false,
    bitNbr:0,
    bitState:0,
    sendAllControls:true,
    cntrlBits:controlBits,
    numIndicationBits:numIndicationBits } );
} // end SendAllControls()

function ClearAllControls() {
  //clear all states in the controlBits array
  for (let i=0; i < controlBits.length; i++) {
    controlBits[i] = 0;
  }

  //Notify the user that all control bits have been cleared
  $( "#message_dialog" ).dialog( "option", "title", "Clear All Controls" );

  document.getElementById("message_dialog").innerHTML = '<div><img style="float: left; margin-top: 5px; width: 36px" src="images/info32px.png"/>'
    + '<div style="float: left; margin-left: 15px; margin-top: 6px; width: 200px">All control bits have been cleared (but not sent).</div></div>';

  ShowMessageDialog();

  //Close the controls dialog box if open.
  if ($( "#controls_dialog" ).dialog( "isOpen" )) {
    $( "#controls_dialog" ).dialog( "close" );
  }

  //console.log(Date() + " Controls: " + controlData);
} // end ClearAllControls()

function ShowMessageDialog() {
  //This function is called to show a message dialog. The innerHTML and title of the dialog
  //are set by the caller
  $( "#message_dialog" ).dialog( "option", "buttons", [{
      text: "Close", click: function() {
        $( this ).dialog( "close" );
      }
    }
  ]);

  $( "#message_dialog" ).dialog( "open" );
} // end ShowMessageDialog()

function pad(n) {
  return (n < 10) ? '0' + n.toString() : n.toString();
} // end pad()

function SecondsElapsed(startTime) {
  let currentTime = new Date();
  let timeDiff = currentTime - startTime; //in ms

  // strip the ms
  timeDiff /= 1000;

  // return seconds
  return Math.round(timeDiff);
} // end SecondsElapsed()

function MillisecondsElapsed(startTime) {
  let currentTime = new Date();
  return currentTime - startTime;
} // end MillisecondsElapsed()

//This function is called when the user clicks on the Open file button
function RequestPanelFileDialog() {
  //Send an http GET request to the QLCP-NET Developer Tool server so that it will provide the Open File Dialog box.
  //Note: Due to browser security, the developer tool must provide the file path to this control panel.
  //If the the request is successful (i.e. there is a response from the developer tool = "open"),
  //set the requestFileName flag so that this web client will periodically
  //send an http request until the user makes a file selection and the file name is received.

  let openrequest = $.get("http://localhost:"+serverPort+"/open/", function(data) {
    if ("open" === data) {
      requestFileName = true;
    }
  }, "text");

  //If the request for Open File Dialog fails, show an error message to the user.
  openrequest.fail(function() {
    ShowDevToolRequestErrorMessage();
  });
} // end RequestPanelFileDialog()

function OpenPanel(file) {

  let doAutoOpen = autoOpen

  //Send an http GET to request the panel file contents
  $.get(file, function(data) {

    let bgImgSz = {};
    let bgImgPos = {};
    let bgImgObj = {};
    let bgImgSrc;

    panelCRC = crc16(data);
    document.getElementById('panel-crc').innerHTML = 'Panel CRC: 0x' + panelCRC.toString(16).toUpperCase().padStart(4, '0');

    //Hide the any open dialog box
    CloseAllDialogBoxes();

    // Remove any component selection in the toolbar.
    DeselectDropComponentButtons();

    //clear the component layer
    component_layer.removeChildren();
    component_layer.clear();

    //remove any background image for the panel (component style designs)
    bgImage_layer.removeChildren();
    bgImage_layer.clear();

    InitGrid();

    // Store each line from the data in an array
    let attrLines = new Array();
    if (data.indexOf("\r") > 0) {
      attrLines = data.split("\r\n");
    } else if (data.indexOf("\n") > 0) {
      attrLines = data.split("\n");
    }

    // Reset id number
    idNbr = 1;

    // Get the version from the first line.
    let compAttr = attrLines.shift().split("=");

    if (+compAttr[1] > CPF_VERSION) {
      $( "#message_dialog" ).dialog( "option", "title", "Control Panel Request" );

      document.getElementById("message_dialog").innerHTML = "<div><img style='float: left; margin-top: 5px; width: 36px' src='images/warning_icon.png'/>"
        + "<div style='float: left; margin-left: 15px; margin-top: 6px; width: 350px'>The panel file being loaded was created with a newer "
        + "version of the QLCP Web GUI SW. Please update the QLCP's Web GUI SW to view this control panel.<br/><br/>"
        + "</div></div>";

      ShowMessageDialog();
      return;
    } else {
      //Request the location information from the QLCP
      $.get("http://" + hostUrl + ":54321/info/", function(data) {
        //Success! Received the location information. Now, display it in the message dialog box.
        $( "#message_dialog" ).dialog( "option", "title", "Control Panel Request" );

        document.getElementById("message_dialog").innerHTML = '<div><img style="float: left; margin-top: 5px; width: 36px" src="images/warning_icon.png"/>'
          + '<div style="float: left; margin-left: 15px; margin-top: 6px; width: 350px">You have requested the control panel for the following location:<br/><br/>'
          + 'IP Address:&nbsp&nbsp' + hostUrl + '<br/>'
          + 'Controller Name:&nbsp&nbsp' + data.split(",")[0] + '<br/>'
          + 'Location Name:&nbsp&nbsp' + data.split(",")[1] + '<br/>'
          + 'Railroad Name:&nbsp&nbsp' + data.split(",")[2] + '<br/>'
          + 'Subdivision Name:&nbsp&nbsp' + data.split(",")[3] + '<br/>'
          + 'Panel CRC: 0x' + panelCRC.toString(16).toUpperCase().padStart(4, '0') + '<br/>'
          + '</div></div>';

        ShowMessageDialog();
      }, "text");
    }

    // Parse each line the panel file.
    while (attrLines.length > 0) {

      // Get the name/value pair (the attribute on this line)
      compAttr = attrLines.shift().split(":");

      // Handle the entry based on the type of object/component it defines.
      switch (compAttr[0]) {

        case "icon_begin":
        {
          alert("Mosaic Tile panels are no longer supported. Please use web-client version 4.11 or earlier..");
          return;
        }
        break;

        // Load an LCP background image.
        case "lcp_background":
        {
          // Create the image object and save the data for use after all params are read.
          bgImgObj = new Image();
          bgImgSrc = compAttr[1] + ":" + compAttr[2];

          bgImgObj.onload = () => {
            let bgImg = new Konva.Image({
              image: bgImgObj,
              x: 20, // default position from previously handled with CSS margin values.
              y: 50,
              stroke: 'rgba(0,255,255,1)',
              strokeWidth: 3,
              strokeEnabled: false,
              draggable: false,
              id: "c" + idNbr++ // dkh tmp...not yet sure how image fits with other components
            });

            let actual_sz = bgImg.size();

            if (bgImgSz.width != undefined) {
              // Maintain original aspect ratio to determine height.
              bgImg.size({
                width: bgImgSz.width,
                height: (bgImgSz.width * (actual_sz.height/actual_sz.width))
              });
            }

            if (bgImgPos != undefined) {
              bgImg.position({
                x: bgImgPos.x,
                y: bgImgPos.y
              });
            }

            bgImage_layer.add(bgImg);
            bgImage_layer.draw();
          }
        }
        break;

        // Store an LCP background image width.
        case "lcp_background_width":
        {
            bgImgSz.width = Number(compAttr[1]);
        }
        break;

        // Store an LCP background image position.
        case "lcp_background_pos":
        {
          bgImgPos.x = Number(compAttr[1].split(',')[0]);
          bgImgPos.y = Number(compAttr[1].split(',')[1]);
        }
        break;

        // Load a single component: read all component attributes, create the component and add it to the layer.
        // Note: "text_begin" and "text_end" is for older panel files.
        case "component_begin":
        case "text_begin":
        {
          let attributes = new Array();
          let indicationAttrs = new Array();
          let controlAttrs = new Array();

          // Split the next line name/value pair
          compAttr = attrLines.shift().split(":");

          // Proccess all attributes until the end of component is detected.
          while ((compAttr[0] !== "component_end") && (compAttr[0] !== "text_end")) {

            let attrName = compAttr[0];
            let attrValue = compAttr[1];

            // Adjust any color values to use new rgba internal representation instead
            // of HTML color strings that were used on older versions.
            if ((attrName == 'fill' )|| (attrName == 'stroke') || (attrName == 'defaultColor')) {

              // Attempt to look up the value by the color name (for older saved panels.)
              let idx = stdColors.indexOf(attrValue);
              if (idx != -1) {
                attrValue = colorVals[idx];
              }
            }

            // Store in the attributes array as a name/value pair.
            attributes[attrName] = attrValue;

            // Search for control attributes
            if (attrName.indexOf('control') >= 0) {
              controlAttrs.push(attrValue);
            }

            // Search for indication attributes
            if (attrName.indexOf('indication') >= 0) {

              // Revert to default color assignments if no color was stored.
              if (attrValue.indexOf('rgba') === -1) {
                // Add a value of true if there was no name assignment to make the string complete.
                if (attrValue.split(",")[0] == '0') {
                  attrValue += ',0';
                }
                attrValue += ',' + colorVals[attrName.replace('indication', '') % (colorVals.length)];
              }
              indicationAttrs.push(attrValue);
            }

            // Get the next line of panel file.
            compAttr = attrLines.shift().split(":");
          } // end while ... (process all attributes)

          // Set the component type to text if it is undefined.
          if (attributes.type === undefined) {
            attributes.type = "text"
          }

          // Create a new component object from the stored attributes.
          dropMode[attributes.type] = true;
          let Component = AddComponent(Math.round(Number(attributes.x)), Math.round(Number(attributes.y)), idNbr);
          dropMode[attributes.type] = false;

          if (Component == null) {
            alert("Error! Unable to create component. Panel file will not be loaded.");
            return;
          }

          // Add the id number to the component attributes and increment the count.
          attributes["id"] = "c" + idNbr++;

          SetComponentProperties(Component, attributes);

          // Set the control properties for this component
          for (let i = 0; i < controlAttrs.length; i++) {
            Component.attrs["control" + i] = controlAttrs[i];
          }

          // Set the indication properties for this component
          for (let i = 0; i < indicationAttrs.length; i++) {
            Component.attrs["indication" + i] = indicationAttrs[i];
          }

          // Update the selection border (needed for older panels that were saved before selection border was updated)
          if (typeof Component.find('Text')[0] !== 'undefined') {
            Component.find('Text')[0].setText(Component.find('Text')[0].attrs.text);
          }
          Component.find('.selection')[0].attrs.update();
        }
        break;
      } // end switch(compAttr[0])
    } // end while (attrLines.length > 0)

    // Now that file parsing is done, load/handle background image.
    if (bgImgObj != undefined) {
      bgImgObj.src = bgImgSrc;
    }

    redrawComponents = true;

    // Go directly into run mode if panel file is being opened as a result of an 'auto open' trigger
    if (doAutoOpen) {
      RunMode();
    }

  }, "text");

  //Clear 'auto open' flag now that panel file has been opened
  autoOpen = false;

} // end OpenPanel()

function SetComponentProperties(Component, attributes) {
  let indicatorObj;
  let textObj;
  let cmpType = Component.attrs.type;

  if (Component.find('.indicator').length > 0) {
    indicatorObj = Component.find('.indicator')[0];
  }

  if (Component.find('Text').length > 0) {
    textObj = Component.find('Text')[0];
  }

  // Set pb_square text properties.
  if (cmpType === "pb_square") {
    if (typeof attributes.text !== 'undefined') {
      textObj.setText(attributes.text);
    }
    if (typeof attributes.fill !== 'undefined') {
      textObj.fill(attributes.fill);
    }
    if (typeof attributes.fontSize !== 'undefined') {
      textObj.fontSize(attributes.fontSize);
    }

    let base = Component.find('Rect')[0];
    let selectRect = Component.find('Rect')[2]

    if (attributes.width !== undefined) {
      indicatorObj.width(+attributes.width);
      base.width(+attributes.width+10);

    }
    if (attributes.height !== undefined) {
      indicatorObj.height(+attributes.height);
      base.height(+attributes.height+10);
    }

    selectRect.width(base.width());
    selectRect.height(base.height());

    textObj.attrs.alignText();  //this has to occur before scaling!
  }

  // Set text or timer text properties.
  if (cmpType === "text" || cmpType === "timer") {

    if (typeof attributes.defaultColor !== 'undefined') {
      indicatorObj.attrs.defaultColor = attributes.defaultColor;
    } else {
      // For older versions that don't support default color.
      indicatorObj.attrs.defaultColor = "rgba(255,255,255,1)";
    }

    if (typeof attributes.fontSize !== 'undefined') {
      textObj.fontSize(attributes.fontSize);
    }

    if (typeof attributes.text !== 'undefined') {
      textObj.setText(attributes.text);
    }

    if(cmpType === "text") {
      if (typeof attributes.align !== 'undefined') {
        textObj.align(attributes.align);
        textObj.attrs.alignText();
      }
    }

  } else if (typeof attributes.scaleX !== 'undefined') {

    //scale applies to all components but text,
    //and yes, scaleX is used for y.
    Component.scale({
      x: attributes.scaleX,
      y: attributes.scaleX
    });
  }

  if (cmpType === "timer") {
    indicatorObj.attrs.timerDuration = attributes.timerDuration;
  }

  // Set LED, quadrilateral, or arrow stroke values.
  if (cmpType === "LED" || cmpType === "quadrilateral" || cmpType === "arrow") {

    if (typeof attributes.stroke !== 'undefined') {
      indicatorObj.stroke(attributes.stroke);
    }

    if (typeof attributes.strokeWidth !== 'undefined') {

      indicatorObj.strokeWidth(attributes.strokeWidth);
      indicatorObj.strokeEnabled((attributes.strokeWidth > 0 ? true: false));
    }

    if (typeof attributes.defaultColor !== 'undefined') {
      indicatorObj.attrs.defaultColor = attributes.defaultColor;
    }
  }

  // Set components with rotation property.
  if (cmpType === "text" || cmpType === "quadrilateral" || cmpType === "arrow" ||
      cmpType === "timer" || cmpType === "togSw" || cmpType === "togSwMom") {

    if (typeof attributes.rotation !== 'undefined') {
      Component.rotation(attributes.rotation);
    }
  }

  // Set quadrilateral point coordinates.
  if (cmpType === "quadrilateral") {

    let quad = indicatorObj.attrs;
    quad.point1x = attributes.point1x;
    quad.point1y = attributes.point1y;
    quad.point2x = attributes.point2x;
    quad.point2y = attributes.point2y;
    quad.point3x = attributes.point3x;
    quad.point3y = attributes.point3y;
    quad.point4x = attributes.point4x;
    quad.point4y = attributes.point4y;
  }

  if (Component.attrs.nbrCtrls > 0) {

    if (typeof attributes.passcode !== 'undefined') {
      Component.attrs.passcode = attributes.passcode;
    }

    if (typeof attributes.timeout !== 'undefined') {
      Component.attrs.timeout = attributes.timeout;
    }
  }

  if (Component.attrs.hasIndications) {
    if (typeof attributes.flashTime !== 'undefined') {
      Component.attrs.flashTime = attributes.flashTime;
      // todo - will I need to assign colors for indicator objects here also?
    }
  }
} // end SetComponentProperties()

function SerializePanelToString() {
    let saveStr = "version = " + CPF_VERSION +"\n";

  //save the text and other components (add their data to the save string)
  let panelObjs = component_layer.getChildren();
  for (let i=0; i < panelObjs.length; i++) {
    saveStr += "component_begin:\n";

    //save all the attributes at the component group level
    for (let key in panelObjs[i].attrs) {
      switch(key)
      {
        case "id":
        case "state":
        case "flashTimeCntr":
        case "flashState":
        case "unlocked":
        case "selected":
        case "draggable":
          break;
        default:
          saveStr += key + ':'+ panelObjs[i].attrs[key] + "\n";
          break;
      }
    }

    //save all the attributes for the component's indicator object
    if (typeof panelObjs[i].find('.indicator')[0] !== 'undefined') {
      for (let key in panelObjs[i].find('.indicator')[0].attrs) {
        switch(key)
        {
          case "radius":
          case "strokeEnabled":
          case "strokeScaleEnabled":
          case "sceneFunc":
          case "alignText":
          case "getBorderValues":
          case "hitFunc":
            break;
          default:
            saveStr += key + ':'+ panelObjs[i].find('.indicator')[0].attrs[key] + "\n";
        }
      }
    }

    //save all the attributes for the component's label object
    if (typeof panelObjs[i].find('.label')[0] !== 'undefined') {
      for (let key in panelObjs[i].find('.label')[0].attrs) {
        switch (key)
        {
          case "alignText":
          case "sceneFunc":
          case "hitFunc":
            break;
          default:
            saveStr += key + ':'+ panelObjs[i].find('.label')[0].attrs[key] + "\n";
        }
      }
    }
    saveStr += "component_end:\n";
  }

  //If it exists, save LCP background image as the base64 string that it is.
  if (bgImage_layer.hasChildren()) {
    let bgImg = bgImage_layer.getChildren()[0];
    saveStr += "lcp_background:" + bgImg.image().src + "\n";
    saveStr += "lcp_background_pos:" + bgImg.x() + "," + bgImg.y() + "\n";
    saveStr += "lcp_background_width:" + bgImg.width() + "\n";
  }

  // Update the calulated CRC and show in the status bar
  panelCRC = crc16(saveStr);
  document.getElementById('panel-crc').innerHTML = 'Panel CRC: 0x' + panelCRC.toString(16).toUpperCase().padStart(4, '0');

  return saveStr;
} // end SerializePanelToString()

//This function is called when the user clicks on the Save button
function SavePanel() {
  let postData = SerializePanelToString();

  //Now, request the QLCP developer tool to open the Save File Dialog box so that the user can select a directory and
  //enter a file name. Send the data to be saved with the POST request. The rest of the save will be
  //handled by the QLCP developer tool when the user clicks on the Save button of the Save File Dialog box.
  let saverequest = $.post("http://localhost:"+serverPort+"/save/", postData, function(){
  }, "text");

  //If the request for Save File Dialog fails, show an error message to the user.
  saverequest.fail(function() {
    ShowDevToolRequestErrorMessage();
  });
} // end SavePanel()

//This function is called when the developer tool is unavailable (no response) during the Open and Save file functions.
function ShowDevToolRequestErrorMessage() {
  $( "#message_dialog" ).dialog( "option", "title", "Request Timeout" );

  document.getElementById("message_dialog").innerHTML = '<div><img style="float: left; margin-top: 5px; width: 36px" src="images/error32px.png"/>'
    + '<div style="float: left; margin-left: 15px; margin-top: 6px; width: 200px">The QLCP-NET Developer Tool is not responding to the request.</div></div>';

  ShowMessageDialog();
} // ShowDevToolRequestErrorMessgae()

//This function shows a dialog box to the user warning that all icons and text will be cleared.
function ShowClearAllDialog() {
  document.getElementById("clear_all_dialog").innerHTML = '<div><img style="float: left; margin-top: 5px; width: 36px" src="images/warning_icon.png"/>'
      + '<div style="float: left; margin-left: 15px; margin-top: 6px; width: 200px">This will delete everything in your current design!</div></div>';

  $( "#clear_all_dialog" ).dialog( "option", "buttons", [
    {
      text: "OK", click: function() {
        $( this ).dialog( "close" );
        ClearAllComponents();
      }
    },
    {
      text: "Cancel", click: function() {
        $( this ).dialog( "close" );
      }
    }
  ]);

  $( "#clear_all_dialog" ).dialog( "open" );
} // ShowClearAllDialog()

function ClearAllComponents() {
  idNbr = 0;

  //clear the components layer
  component_layer.removeChildren();
  component_layer.clear();

  //remove background image for the panel if it exists
  if (document.getElementById('lcp_background').childElementCount > 0) {
    document.getElementById('lcp_background').removeChild(document.getElementById('lcp_background').firstChild);
  }

  //initialize the grid and return to design mode
  InitGrid();
  DesignMode();

  if ($( "#properties_dialog" ).dialog( "isOpen" )) {
    $( "#properties_dialog" ).dialog( "close" );
  }
} // end ClearAllComponents()

function ToggleDropComponentButtons(id) {
  // Toggle the drop mode state and show this state by
  // adding or removing the button down class.
  if (dropMode[buttonToComponent[id]]) {
    DeselectDropComponentButtons();
  } else {
    DeselectDropComponentButtons();
    $( '#' + id ).addClass("buttonDown");
    dropMode[buttonToComponent[id]] = true;
    dropMode["none"] = false;
  }
} // end ToggleDropComponentButtons()

function DeselectDropComponentButtons() {
  //clear the values for all of the component buttons
  for (let key in buttonToComponent) {
    $( '#' + key ).removeClass("buttonDown");
  }

  //clear all modes
  for (let key in dropMode) {
    dropMode[key] = false;
  }

  //set mode to none
  dropMode["none"] = true;
} // end DeselectDropComponentButtons()

//********************* Component functions/classes ****************************
//******************************************************************************
// All components are groups of shapes that inherit the common properties and
// methods of the group created in this function.
function ComponentGroup(xpos, ypos, id_nbr, cmp_type, cmp_text, numCtrls, has_indications) {
  let Component = new Konva.Group({
    x: xpos,
    y: ypos,
    scaleX: 1,
    scaleY: 1,
    offsetX: 0,
    offsetY: 0,
    skewX: 0,
    skewY: 0,
    type: cmp_type,
    text: cmp_text,
    nbrCtrls: numCtrls,
    hasIndications: has_indications,
    state: false,
    rotation: 0,
    flashTime: 500,
    flashTimeCntr: 0,
    flashState: 1,
    passcode: '',
    unlocked: false,
    timeout: 0,
    selected: false,
    draggable: (TogglePositionLock.positionLocked == true ? false : true),
    releaseTimeout: 0,
    id: "c" + id_nbr
  });

  return Component;
} // end ComponentGroup()

//All components inherit the selection border rectangle and sizing handles in this function
function ComponentSelectionBorder() {
  let selectionBorder = new Konva.Rect({
    name: "selection",
    fill: "rgba(0,0,0,0)",
    stroke: "rgba(0,255,255,1)",
    strokeWidth: 2,
    strokeEnabled: false,

    update: function()
    {
      if (selectionBorder.getParent().attrs.type === 'quadrilateral' ||
          selectionBorder.getParent().attrs.type === 'LED' ||
          selectionBorder.getParent().attrs.type === 'text' ||
          selectionBorder.getParent().attrs.type === 'timer')
      {
        let component = selectionBorder.getParent();
        let borderValues = component.find('.indicator')[0].attrs.getBorderValues();
        selectionBorder.position({
          x: borderValues.x,
          y: borderValues.y
        });
        selectionBorder.attrs.width = borderValues.width;
        selectionBorder.attrs.height = borderValues.height;
      }
    },

    show: function()
    {
      selectionBorder.strokeEnabled(true);
    },

    hide: function()
    {
      selectionBorder.strokeEnabled(false);
    }
  });

  return selectionBorder;
} // end ComponentSelectionBorder()

//This function creates a new quadrilateral.
function CreateQuadrilateral(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'quadrilateral', '', 0, true);

  //create a quadrilateral shape (rectangle initially)
  let quad = new Konva.Shape({
    name: 'indicator', //this is the shape that shows the indications for the component
    fill: defaultFillColor,
    defaultColor: defaultFillColor,
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 0,
    strokeEnabled: false,
    point1x: 0,
    point1y: 0,
    point2x: 70,
    point2y: 0,
    point3x: 70,
    point3y: 18,
    point4x: 0,
    point4y: 18,

    // a Konva.Canvas renderer is passed into the sceneFunc function
    sceneFunc: function(context, shape)
    {
      context.beginPath();
      context.moveTo(this.attrs.point1x, this.attrs.point1y);
      context.lineTo(this.attrs.point2x, this.attrs.point2y);
      context.lineTo(this.attrs.point3x, this.attrs.point3y);
      context.lineTo(this.attrs.point4x, this.attrs.point4y);
      context.closePath();
      context.fillStrokeShape(shape);
    },

    getBorderValues: function()
    {
      let x, y, width, height;

      x = -quad.attrs.strokeWidth/Component.attrs.scaleX/2;
      y = -quad.attrs.strokeWidth/Component.attrs.scaleX/2;

      //for width, use the largest x value
      if (quad.attrs.point2x > quad.attrs.point3x)
      {
        width = Number(quad.attrs.point2x) + (quad.attrs.strokeWidth/Component.attrs.scaleX);
      }
      else
      {
        width = Number(quad.attrs.point3x) + (quad.attrs.strokeWidth/Component.attrs.scaleX);
      }

      //height, use the largest y value
      if (quad.attrs.point3y > quad.attrs.point4y)
      {
        height = Number(quad.attrs.point3y) + (quad.attrs.strokeWidth/Component.attrs.scaleX);
      }
      else
      {
        height = Number(quad.attrs.point4y) + (quad.attrs.strokeWidth/Component.attrs.scaleX);
      }

      return {
        x: x,
        y: y,
        width: width,
        height: height
      };
    }
  });

  //we don't want the stroke width to scale
  quad.strokeScaleEnabled(false);

  //create a selection border for this component and set its initial width and height
  let selectionBorder = ComponentSelectionBorder();
  let borderValues = quad.attrs.getBorderValues();
  selectionBorder.position({
    x: borderValues.x,
    y: borderValues.y
  });
  selectionBorder.attrs.width = borderValues.width;
  selectionBorder.attrs.height = borderValues.height;
  selectionBorder.strokeScaleEnabled(false);

  //add shapes to the group
  Component.add(quad, selectionBorder);

  //mouse down event
  Component.on('mousedown', function() {
    ComponentMouseDown(this);
  });

  //create a reference for the message listener to be used when we delete this component
  Component.msgListener = function(e){
    UpdateIndication(e.data.indication, e.data.indicationMask, Component, quad);
  };

  //add listener for message from web worker
  QnetClientThread.addEventListener("message", Component.msgListener, false);

  return Component;
} // end CreateQuadrilateral()

//This function creates a new arrow.
function CreateArrow(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'arrow', '', 0, true);

  //create an arrow
  let arrow = new Konva.Shape({
    name: "indicator",
    fill: defaultFillColor,
    defaultColor: defaultFillColor,
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 0,
    strokeEnabled: false,

    // a Konva.Canvas renderer is passed into the sceneFunc function
    sceneFunc: function(context, shape)
    {
      context.beginPath();
      context.moveTo(0,0);
      context.lineTo(20,0);
      context.lineTo(20,-5);
      context.lineTo(40,5);
      context.lineTo(20,15);
      context.lineTo(20,10);
      context.lineTo(0,10);
      context.closePath();
      context.fillStrokeShape(shape);
    }
  });

  arrow.strokeScaleEnabled(false);

  //create a highlight for this component that inherits the common component highlight properties
  let selectionBorder = ComponentSelectionBorder();
  selectionBorder.attrs.y = -5;
  selectionBorder.attrs.width = 40;
  selectionBorder.attrs.height = 20;
  selectionBorder.strokeScaleEnabled(false);

  //add shapes to the group
  Component.add(arrow, selectionBorder);

  //mouse down event
  Component.on('mousedown', function() {
    ComponentMouseDown(this);
  });

  //create a reference for the message listener to be used when we delete this component
  Component.msgListener = function(e){
    UpdateIndication(e.data.indication, e.data.indicationMask, Component, arrow);
  };

  //add listener for message from web worker
  QnetClientThread.addEventListener("message", Component.msgListener, false);

  return Component;
} // end CreateArrow()

//This function creates a new LED.
function CreateLED(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'LED', '', 0, true);

  //create the led shape
  let led = new Konva.Circle({
    name: "indicator",
    radius: 12,
    fill: defaultFillColor,
    defaultColor: defaultFillColor,
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 2,
    strokeEnabled: true,

    getBorderValues: function()
    {
      let x, y, width, height;

      x = -(Number(led.attrs.radius) + led.attrs.strokeWidth/Component.attrs.scaleX/2);
      y = -(Number(led.attrs.radius) + led.attrs.strokeWidth/Component.attrs.scaleX/2);
      width = 2 * Number(led.attrs.radius) + led.attrs.strokeWidth/Component.attrs.scaleX;
      height = width;

      return {
        x: x,
        y: y,
        width: width,
        height: height
      };
    }
  });

  led.strokeScaleEnabled(false);
  //led.perfectDrawEnabled(false); //this could help performance

  //create a selection border for this component and set its initial position, width, and height
  let selectionBorder = ComponentSelectionBorder();
  let borderValues = led.attrs.getBorderValues();
  selectionBorder.position({
    x: borderValues.x,
    y: borderValues.y
  });
  selectionBorder.attrs.width = borderValues.width;
  selectionBorder.attrs.height = borderValues.height;
  selectionBorder.strokeScaleEnabled(false);

  //add shapes to the group
  Component.add(led, selectionBorder);

  //mouse down event
  Component.on('mousedown', function() {
    ComponentMouseDown(this);
  });

  //create a reference for the message listener to be used when we delete this component
  Component.msgListener = function(e){
    UpdateIndication(e.data.indication, e.data.indicationMask, Component, led);
  };

  //add listener for message from web worker
  QnetClientThread.addEventListener("message", Component.msgListener, false);

  return Component;
} // end CreateLED()

//This function creates a new pushbutton.
function CreatePushbutton(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'pb', '', 2, true);

  let swBaseRing = new Konva.Circle({
    radius: 15,
    fill: "rgba(211,211,211,1)",
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1
  });

  let swButton = new Konva.Circle({
    name: "indicator",
    radius: 9,
    fill: defaultFillColor,
    defaultColor: defaultFillColor,
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 2
  });

  //create a selection border for this component and initialize its values
  let selectionBorder = ComponentSelectionBorder();
  selectionBorder.attrs.x = -swBaseRing.attrs.radius;
  selectionBorder.attrs.y = -swBaseRing.attrs.radius;
  selectionBorder.attrs.width = swBaseRing.attrs.radius * 2;
  selectionBorder.attrs.height = swBaseRing.attrs.radius * 2;
  selectionBorder.strokeScaleEnabled(false);

  Component.add(swBaseRing, swButton, selectionBorder);

  Component.addEventListener('mousedown', function() {rndBtnPressedHandler(this);});
  Component.addEventListener('touchstart', function(e) {rndBtnPressedHandler(this); e.stopPropagation();});

  let rndBtnPressedHandler = function(cmp) {

    ComponentMouseDown(cmp);

    if (designMode === false) {
      CheckLockStatus(cmp);
      lastClickTime = new Date();

      if (cmp.attrs.unlocked) {
        swButton.attrs.radius = 8;
        swButton.attrs.strokeWidth = 3;
        redrawComponents = true;
        SendControl(cmp.attrs["control0"]-1, 1);
      }

      // Set a timeout for touch events if the user does not release or moves/slides off the button.
      cmp.releaseTimeout = setTimeout(function(){rndBtnReleasedHandler(cmp);}, 1000);
    }
  };

  //Add button released event listeners.
  //If it is a touch event, we don't want the mouse event to be processed so stop propagation/bubbling.
  //(touch events occur first)
  Component.addEventListener('mouseup', function() {rndBtnReleasedHandler(this);});
  Component.addEventListener('mouseleave', function() {rndBtnReleasedHandler(this);});
  Component.addEventListener('touchend', function(e) {rndBtnReleasedHandler(this); e.stopPropagation();});

  let rndBtnReleasedHandler = function (cmp) {

    if (designMode === false && cmp.attrs.unlocked)
    {
      swButton.attrs.radius = 9;
      swButton.attrs.strokeWidth = 2;
      redrawComponents = true;
      SendControl(cmp.attrs["control0"]-1, 0);

      // Cancel the release timeout
      clearTimeout(cmp.releaseTimeout);
    }
  };

  //create a reference for the message listener to be used when we delete this component
  Component.msgListener = function(e){
    UpdateIndication(e.data.indication, e.data.indicationMask, Component, swButton);
  };

  //add listener for message from web worker
  QnetClientThread.addEventListener("message", Component.msgListener, false);

  return Component;
} // end CreatePushbutton()

//This function creates a new square pushbutton.
function CreateSquarePushbutton(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'pb_square', '', 2, true);

  let swBase = new Konva.Rect({
    width: 65,
    height: 30,
    fill: defaultFillColor,
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1
  });

  let swButton = new Konva.Rect({
    name: "indicator",
    offsetX: -5,
    offsetY: -5,
    width: 55,
    height: 20,
    fill: "rgba(0,0,0,1)",
    defaultColor: "rgba(0,0,0,1)",
    stroke: "rgba(211,211,211,1)",
    strokeWidth: 2
  });

  let swText = new Konva.Text({
    name: "label",
    text: '',
    fontSize: 10,
    fontFamily: 'Arial',
    fill: "rgba(255,255,255,1)",
    alignText: function()
    {
      swText.offsetX(-swBase.width()/2 + swText.width()/2);
      swText.offsetY(-swBase.height()/2 + swText.height()/2);
    }
  });

  //create a selection border for this component and initialize its values
  let selectionBorder = ComponentSelectionBorder();
  selectionBorder.attrs.width = swBase.getWidth();
  selectionBorder.attrs.height = swBase.getHeight();
  selectionBorder.strokeScaleEnabled(false);

  Component.add(swBase, swButton, swText, selectionBorder);

  //Add button pressed event listeners.
  //If it is a touch event, we don't want the mouse event to be processed also so stop propagation/bubbling.
  //(touch events occur first)
  Component.addEventListener('mousedown', function() {btnPressedHandler(this);});
  Component.addEventListener('touchstart', function(e) {btnPressedHandler(this); e.stopPropagation();});

  let btnPressedHandler = function (cmp) {
    ComponentMouseDown(cmp);
    if (designMode === false)
    {
      CheckLockStatus(cmp);
      lastClickTime = new Date();

      if (cmp.attrs.unlocked)
      {
        swButton.width(swBase.width()-12);
        swButton.height(swBase.height()-12);
        swButton.attrs.strokeWidth = 3;
        swButton.offsetX(-6);
        swButton.offsetY(-6);
        swBase.attrs.fill = "rgba(255,0,0,1)";
        redrawComponents = true;
        SendControl(cmp.attrs["control0"]-1, 1);
      }

      //Set a timeout for this button if the user does not release or moves/slides off the button.
      //This is needed for touch events.
      cmp.releaseTimeout = setTimeout(function(){btnReleasedHandler(cmp);}, 1000);
    }
  };

  //Add button released event listeners.
  //If it is a touch event, we don't want the mouse event to be processed so stop propagation/bubbling.
  //(touch events occur first)
  Component.addEventListener('mouseup', function() {btnReleasedHandler(this);});
  Component.addEventListener('mouseleave', function() {btnReleasedHandler(this);});
  Component.addEventListener('touchend', function(e) {btnReleasedHandler(this); e.stopPropagation();});

  let btnReleasedHandler = function (cmp) {
    if (designMode === false && cmp.attrs.unlocked)
    {
      swButton.width(swBase.width()-10);
      swButton.height(swBase.height()-10);
      swButton.attrs.strokeWidth = 2;
      swButton.offsetX(-5);
      swButton.offsetY(-5);
      swBase.attrs.fill = defaultFillColor;
      redrawComponents = true;
      SendControl(cmp.attrs["control0"]-1, 0);

      //cancel the release timeout
      clearTimeout(cmp.releaseTimeout);
    }
  };

  //create a reference for the message listener to be used when we delete this component
  Component.msgListener = function(e){
    UpdateIndication(e.data.indication, e.data.indicationMask, Component, swButton);
  };

  //add listener for message from web worker
  QnetClientThread.addEventListener("message", Component.msgListener, false);

  return Component;
} // end CreateSquarePushbutton()

//This function creates a new toggle switch.
function CreateToggleSwitch(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'togSw', '', 3, false);

  let swBaseRing = new Konva.Circle({
    radius: 17,
    fill: "rgba(211,211,211,1)",
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1
  });

  let swBaseCntr = new Konva.Circle({
    radius: 7,
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1
  });

  let switchUp = new Konva.Shape({
    name: "up",
    fill: "rgba(211,211,211,1)",
    stroke: "rgba(0,0,0,1)",
    visible: false,
    strokeWidth: 1,
    // a Konva.Canvas renderer is passed into the sceneFunc function
    sceneFunc: function(context, shape)
    {
      context.beginPath();
      context.moveTo(-5, 0);
      context.lineTo(-12, -25);
      context.lineTo(12, -25);
      context.lineTo(5, 0);
      context.bezierCurveTo(5,7,-5,7,-5,0);
      context.moveTo(-12, -25);
      context.lineTo(-12, -30);
      context.lineTo(12,-30);
      context.lineTo(12,-25);
      context.fillStrokeShape(shape);
    }
  });

  let switchDown = switchUp.clone();
  switchDown.name('down');
  switchDown.rotate(180);

  let switchCenter = new Konva.Shape({
    name: "center",
    fill: "rgba(211,211,211,1)",
    stroke: "rgba(0,0,0,1)",
    visible: false,
    strokeWidth: 1,
    // a Konva.Canvas renderer is passed into the sceneFunc function
    sceneFunc: function(context, shape)
    {
      context.beginPath();
      context.moveTo(-12, 5);
      context.lineTo(-12, -5);
      context.lineTo(12, -5);
      context.lineTo(12, 5);
      context.lineTo(-12, 5);
      context.fillStrokeShape(shape);
    }
  });

  //create a selection border for this component and initialize its values
  let selectionBorder = ComponentSelectionBorder();
  selectionBorder.attrs.x = -swBaseRing.attrs.radius;
  selectionBorder.attrs.y = -swBaseRing.attrs.radius;
  selectionBorder.attrs.width = swBaseRing.attrs.radius * 2;
  selectionBorder.attrs.height = swBaseRing.attrs.radius * 2;
  selectionBorder.strokeScaleEnabled(false);

  // Add the component pieces to the group.
  Component.add(swBaseRing, swBaseCntr, switchUp, switchDown, switchCenter, selectionBorder);

  // When adding component from saved file...
  // If this is a momentary toggle switch, change the type
  if (dropMode.togSwMom) {
    Component.attrs.type = "togSwMom";
    switchCenter.visible(true);
  } else {
    switchDown.visible(true);
  }

  Component.addEventListener('mousedown', function() {togglePressedHandler(this);});
  Component.addEventListener('touchstart', function(e) {togglePressedHandler(this); e.stopPropagation();});

  let togglePressedHandler = function(cmp) {

    ComponentMouseDown(cmp);

    if (designMode === false || initLatchedSwitches === true) {
      //e.cancelBubble = true; //no need to bubble this event
      CheckLockStatus(cmp);
      lastClickTime = new Date();

      if ((cmp.attrs.unlocked ||initLatchedSwitches === true) && cmp.attrs.type === "togSw") {
        if (initLatchedSwitches) {
          // Also clear (un)lock bit.
          controlBits[cmp.attrs["control2"]-1] = 0;
        }

        if (cmp.attrs.state) {
          //turn the switch off
          cmp.attrs.state = false;
          switchUp.visible(false);
          switchDown.visible(true);
          controlBits[cmp.attrs["control0"]-1] = 0;
          controlBits[cmp.attrs["control1"]-1] = 1;
          // if (initLatchedSwitches === false) {
            SendAllControls();
          // }
        } else {
          cmp.attrs.state = true;
          switchUp.visible(true);
          switchDown.visible(false);
          controlBits[cmp.attrs["control0"]-1] = 1;
          controlBits[cmp.attrs["control1"]-1] = 0;
          // if (initLatchedSwitches === false) {
            SendAllControls();
          // }
        }
        redrawComponents = true;
      }
      if ((cmp.attrs.unlocked || initLatchedSwitches) && cmp.attrs.type === "togSwMom") {
        //If not rotated and clicked on upper side of switch,
        //Or if rotated and clicked on right side of switch, send control0.
        if (cmp.attrs.rotation == 0 && (Math.round(stage.getPointerPosition().y) < (cmp.y() - scrollContainer.scrollTop)) ||
            cmp.attrs.rotation == 90 && (Math.round(stage.getPointerPosition().x) > (cmp.x() - scrollContainer.scrollLeft))) {
          switchUp.visible(true);
          switchDown.visible(false);
          switchCenter.visible(false);
          SendControl(cmp.attrs["control0"]-1, 1);
        } else { //If clicked on lower side of switch, send control1
          switchUp.visible(false);
          switchDown.visible(true);
          switchCenter.visible(false);
          SendControl(cmp.attrs["control1"]-1, 1);
        }
        redrawComponents = true;
      }
    }
  };

  Component.addEventListener('mouseup', function() {toggleReleasedHandler(this);});
  Component.addEventListener('mouseleave', function() {toggleReleasedHandler(this);});
  Component.addEventListener('touchend', function(e) {toggleReleasedHandler(this); e.stopPropagation();});

  let toggleReleasedHandler = function (cmp) {

    if (designMode === false && cmp.attrs.unlocked && cmp.attrs.type === "togSwMom") {
      switchUp.visible(false);
      switchDown.visible(false);
      switchCenter.visible(true);
      redrawComponents = true;
      SendControl(cmp.attrs["control0"]-1, 0);
      SendControl(cmp.attrs["control1"]-1, 0);
    }
  };

  return Component;
} // end CreateToggleSwitch()

//This function creates a new timer component.
function CreateTimer(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'timer', '', 0, true);

  //create a new text object
  let timer = new Konva.Text({
    name: 'indicator',
    text: "0s",
    timerDuration: 0,
    startTime: 0,
    timerRunning: false,
    timerComplete: false,
    fontSize: 20,
    fontFamily: 'Arial',
    fill: "rgba(255,255,255,1)",
    defaultColor: "rgba(255,255,255,1)",

    getBorderValues: function()
    {
      let x, y, width, height;

      x = 0;
      y = 0;
      width = timer.getWidth();
      height = timer.getHeight();

      return {
        x: x,
        y: y,
        width: width,
        height: height
      };
    }
  });

  //create a selection border for this component and initialize its values
  let selectionBorder = ComponentSelectionBorder();
  selectionBorder.attrs.width = timer.getWidth();
  selectionBorder.attrs.height = timer.getHeight();
  selectionBorder.strokeScaleEnabled(false);

  // Add the component pieces to the group.
  Component.add(timer, selectionBorder);

  //mouse down event
  Component.on('mousedown', function() {
    ComponentMouseDown(this);
  });

  //create a reference for the message listener to be used when we delete this component
  Component.msgListener = function(e){
    UpdateIndication(e.data.indication, e.data.indicationMask, Component, timer);
  };

  //add listener for message from web worker
  QnetClientThread.addEventListener("message", Component.msgListener, false);

  return Component;
} // end CreateTimer()

//This function creates a new text component.
function CreateText(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'text', '', 0, true);

  //create a new text object
  let cmpText = new Konva.Text({
    name: 'indicator',
    text: 'text',
    fontSize: 14,
    fontFamily: 'Arial',
    fill: "rgba(255,0,0,1)",
    defaultColor: "rgba(255,0,0,1)",
    align: lastTextAlign,

    alignText: function()
    {
      switch (cmpText.align()) {
        case "left":
            cmpText.getParent().offsetX(0);
            break;
        case "center":
            cmpText.getParent().offsetX(cmpText.getWidth()/2);
        break;
        case "right":
            cmpText.getParent().offsetX(cmpText.getWidth());
        break;
      }
    },

    getBorderValues: function()
    {
      let x, y, width, height;

      x = 0;
      y = 0;
      width = cmpText.getWidth();
      height = cmpText.getHeight();

      return {
        x: x,
        y: y,
        width: width,
        height: height
      };
    }
  });

  //create a selection border for this component and initialize its values
  let selectionBorder = ComponentSelectionBorder();
  selectionBorder.attrs.width = cmpText.getWidth();
  selectionBorder.attrs.height = cmpText.getHeight();
  selectionBorder.strokeScaleEnabled(false);

  // Add the component pieces to the group.
  Component.add(cmpText, selectionBorder);

  //mouse down event
  Component.on('mousedown', function() {
    ComponentMouseDown(this);
  });

  //create a reference for the message listener to be used when we delete this component
  Component.msgListener = function(e){
    UpdateIndication(e.data.indication, e.data.indicationMask, Component, cmpText);
  };

  //add listener for message from web worker
  QnetClientThread.addEventListener("message", Component.msgListener, false);

  return Component;
} // end CreateText()

//This function creates a new keyswitch.
function CreateKeySwitch(xpos, ypos, id_nbr) {
  //create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'keyswitch', '', 4, false);

  let swBaseRing = new Konva.Circle({
    radius: 25,
    fill: "rgba(211,211,211,1)",
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1
  });

  let swBaseInnerRing = new Konva.Circle({
    radius: 20,
    fill: "rgba(211,211,211,1)",
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1
  });

  //up and initial position
  let position1 = new Konva.Rect({
    name: "pos1",
    fill: "rgba(0,0,0,1)",
    visible: true,
    strokeWidth: 0,
    x: -4,
    y: -20,
    width: 8,
    height: 25
  });

  //right position
  let position2 = new Konva.Rect({
    name: "pos2",
    fill: "rgba(0,0,0,1)",
    visible: false,
    strokeWidth: 0,
    x: -5,
    y: -4,
    width: 25,
    height: 8
  });

  //left position
  let position3 = new Konva.Rect({
    name: "pos3",
    fill: "rgba(0,0,0,1)",
    visible: false,
    strokeWidth: 0,
    x: -20,
    y: -4,
    width: 25,
    height: 8,
  });

  //create a selection border for this component and initialize its values
  let selectionBorder = ComponentSelectionBorder();
  selectionBorder.attrs.x = -swBaseRing.attrs.radius;
  selectionBorder.attrs.y = -swBaseRing.attrs.radius;
  selectionBorder.attrs.width = swBaseRing.attrs.radius * 2;
  selectionBorder.attrs.height = swBaseRing.attrs.radius * 2;
  selectionBorder.strokeScaleEnabled(false);

  // Add the component pieces to the group.
  Component.add(swBaseRing, swBaseInnerRing, position1, position2, position3, selectionBorder);

  Component.addEventListener('mousedown', function() {ksPressedEventHandler(this);});
  Component.addEventListener('touchstart', function(e) {ksPressedEventHandler(this); e.stopPropagation();});

  let ksPressedEventHandler = function(cmp) {

    ComponentMouseDown(cmp);
    if (designMode === false || initLatchedSwitches === true)
    {
      CheckLockStatus(cmp);
      lastClickTime = new Date();

      if (cmp.attrs.unlocked || initLatchedSwitches)
      {
        if (initLatchedSwitches)
        {
          //turn switch to position 1 (up)
          cmp.attrs.state = false;
          position1.visible(true);
          position2.visible(false);
          position3.visible(false);
          controlBits[cmp.attrs["control0"]-1] = 1;
          controlBits[cmp.attrs["control1"]-1] = 0;
          controlBits[cmp.attrs["control2"]-1] = 0;
          // Also clear (un)lock bit.
          controlBits[cmp.attrs["control3"]-1] = 0;
          SendAllControls();
        }
        //If clicked on upper side of switch or initializing switch position...
        else if (Math.round(stage.getPointerPosition().y) < (cmp.y() - scrollContainer.scrollTop - 5))
        {
          //turn switch to position 1 (up)
          cmp.attrs.state = false;
          position1.visible(true);
          position2.visible(false);
          position3.visible(false);
          controlBits[cmp.attrs["control0"]-1] = 1;
          controlBits[cmp.attrs["control1"]-1] = 0;
          controlBits[cmp.attrs["control2"]-1] = 0;
          SendAllControls();
        }
        //if the click was on the right side of the switch
        else if (Math.round(stage.getPointerPosition().x) > (cmp.x() - scrollContainer.scrollLeft))
        {
          //turn switch to position 2 (right)
          cmp.attrs.state = true;
          position1.visible(false);
          position2.visible(true);
          position3.visible(false);
          controlBits[cmp.attrs["control0"]-1] = 0;
          controlBits[cmp.attrs["control1"]-1] = 1;
          controlBits[cmp.attrs["control2"]-1] = 0;
          SendAllControls();
        }
        //if the click was on the left and the bit for this control has been assigned...
        else if (Math.round(stage.getPointerPosition().x) < (cmp.x() - scrollContainer.scrollLeft) && cmp.attrs["control2"] > 0)
        {
          //turn switch to position 3 (left)
          cmp.attrs.state = true;
          position1.visible(false);
          position2.visible(false);
          position3.visible(true);
          controlBits[cmp.attrs["control0"]-1] = 0;
          controlBits[cmp.attrs["control1"]-1] = 0;
          controlBits[cmp.attrs["control2"]-1] = 1;
          SendAllControls();
        }

        redrawComponents = true;
      }
    }
  };

  return Component;
} // end CreateKeySwitch()

//This function creates a new rotary switch.
function CreateRotarySwitch(xpos, ypos, id_nbr) {
  // create a component that inherits the group component properties
  let Component = ComponentGroup(xpos, ypos, id_nbr, 'rotaryswitch', '', 7, false);

  let dial = new Konva.Shape({
    name: "dial",
    x: 0,
    y: -2.5,
    fill: "rgba(211,211,211,1)",
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1,

    // a Konva.Context renderer is passed into the sceneFunc function
    sceneFunc: function(context, shape) {
      context.beginPath();
      context.moveTo(-10,-8);
      context.lineTo(0,-28);
      context.lineTo(10,-8);
      context.closePath();
      context.fillStrokeShape(shape);
    }
  });

  let centerline = new Konva.Line({
    points: [0, -30, 0, 0],
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1,
  });

  let rswBaseRing = new Konva.Circle({
    radius: 20,
    fill: "rgba(169,169,169,1)",
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1,
  });

  let rswBaseInnerRing = new Konva.Circle({
    radius: 15,
    fill: "rgba(211,211,211,1)",
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1,
    // strokeEnabled: false,
  });

  //up and initial position
  let dialbase = new Konva.Rect({
    name: "dialbase",
    fill: "rgba(211,211,211,1)",
    visible: true,
    stroke: "rgba(0,0,0,1)",
    strokeWidth: 1,
    x: -10,
    y: 10,
    width: 20,
    height: 18
  });

  Component.detent = 0;
  //create a selection border for this component and initialize its values
  let selectionBorder = ComponentSelectionBorder();
  selectionBorder.attrs.x = -rswBaseRing.attrs.radius - 2;
  selectionBorder.attrs.y = -rswBaseRing.attrs.radius - 12;
  selectionBorder.attrs.width = (rswBaseRing.attrs.radius * 2)+4;
  selectionBorder.attrs.height = (rswBaseRing.attrs.radius * 2) + 22;
  selectionBorder.strokeScaleEnabled(false);

  // Add the component pieces to the group.
  Component.add(rswBaseRing, dial, dialbase, rswBaseInnerRing, centerline, selectionBorder);

  Component.addEventListener('mousedown', function() {rsPressedEventHandler(this);});
  Component.addEventListener('touchstart', function(e) {rsPressedEventHandler(this); e.stopPropagation();});

  let rsPressedEventHandler = function(cmp) {

    ComponentMouseDown(cmp);

    if (designMode === false || initLatchedSwitches === true) {
      CheckLockStatus(cmp);
      lastClickTime = new Date();

      if (cmp.attrs.unlocked || initLatchedSwitches) {
        let yPos = Math.round(stage.getPointerPosition().y)
        if (initLatchedSwitches)
        {
          // turn switch to position 1 (up)
          cmp.detent = 0;
          cmp.rotation(0);
          // Clear the lock bit
          controlBits[cmp.attrs["control"+(cmp.attrs.nbrCtrls-1)]-1] = 0;
        }
        //If clicked on upper side of switch or initializing switch position...
        else if (yPos < (cmp.y() - scrollContainer.scrollTop))
        {
          cmp.detent = (cmp.detent == 0) ? 5 : (cmp.detent - 1);
          cmp.rotation(cmp.detent * 60);
        }
        else {
          cmp.detent = (cmp.detent + 1) % 6;
          cmp.rotation(cmp.detent * 60);
        }

        if (cmp.detent === 0) {
          cmp.attrs.state = false;
        } else {
          cmp.attrs.state = true;
        }

        // Loop through all controls (except for the last one / lock bit.)
        for (let ctrlnum = 0; ctrlnum < cmp.attrs.nbrCtrls-1; ctrlnum++) {

          controlBits[cmp.attrs["control"+ctrlnum]-1] = 0;

          if(ctrlnum === cmp.detent) {
            controlBits[cmp.attrs["control"+ctrlnum]-1] = 1;
          }
        }
        SendAllControls();
        redrawComponents = true;
      } // end if (unocked | init)
    }
  };

  return Component;
} // end CreateRotarySwitch()

function ComponentMouseDown(component) {
  //if in developer mode...
  if (developerMode) {
    if (designMode) {
      ShowSelectedComponent(component);
      mouseDownOnComponent = true;
    }
  }
} // end ComponentMouseDown()

//This function updates the component indications and sets the redrawComponents flag if any indication has changed
function UpdateIndication(indicationBits, indicationMaskBits, componentObj, indicator) {
  let bit;
  let state = 0;
  let mask = 1;                     // by default, bit state is considered certain
  let timerComponent;
  let timerState = 0;
  let color1 = indicator.attrs.defaultColor;

  //check the first group of colored indication statuses
  if (designMode === false) {
    // Update the flash counter.
    componentObj.attrs.flashTimeCntr++;

    //invert the state if flash time has elapsed. Timers are 100 ms resolution
    if (componentObj.attrs.flashTimeCntr * 100 >= componentObj.attrs.flashTime) {
      componentObj.attrs.flashTimeCntr = 0;
      componentObj.attrs.flashState = !componentObj.attrs.flashState;
    }

    //set color based on the indication bits
    let indNbr = 0;
    for (let i=0; i < stdColors.length; i++) {
      let indStr = componentObj.attrs["indication" + indNbr];
      if (indStr.split(",")[0] > 0) {
        //get the bit number and state
        bit = indStr.split(",")[0] - 1;

        state = indicationBits[bit];
        mask = indicationMaskBits[bit];

        //invert the state if state option is set to false
        if (1 === Number(indStr.split(",")[1])) {
          state = !state;
        }

        //set the timer state
        if (componentObj.attrs.type === "timer") {
          timerState = timerState | state;
        }

        //set the fill color based on the state value
        if (state) {
          color1 = indStr.indexOf(',');
          color1 = indStr.indexOf(',', color1+1);
          color1 = indStr.slice(color1+1);
        }
      }
      indNbr++;
    }

    //set color based on the flashing indication bits
    for (let i=0; i < stdColors.length; i++) {
      let indStr = componentObj.attrs["indication" + indNbr];
      if (indStr.split(",")[0] > 0) {
        //get the bit number and state
        bit = indStr.split(",")[0] - 1;

        state = indicationBits[bit];

        //invert the state if state option is set to false
        if (1 === Number(indStr.split(",")[1])) {
          state = !state;
        }

        //set the timer state
        if (componentObj.attrs.type === "timer") {
          timerState = timerState | state;
        }

        state &= componentObj.attrs.flashState;
        //console.log("state = " + state);

        if (state) {
          color1 = indStr.indexOf(',');
          color1 = indStr.indexOf(',', color1+1);
          color1 = indStr.slice(color1+1);
        }
      }
      indNbr++;
    }

    //if this is a timer component, set the timer values based on the state value
    if (componentObj.attrs.type === "timer") {
      timerComponent = indicator.attrs;

      if (timerState === 1) {
        if (timerComponent.timerRunning === false && timerComponent.timerComplete === false) {
          timerComponent.timerRunning = true;
          timerComponent.startTime = new Date();
        }
      } else {
        //This means that there are no indication bits associated with this timer that are true.
        timerComponent.timerRunning = false;
        timerComponent.timerComplete = false;
      }
    }

    //Update the timer component's displayed time value
    if (componentObj.attrs.type === "timer") {
      timerComponent = indicator.attrs;

      if(timerComponent.timerRunning) {
        let elapsedTime = SecondsElapsed(timerComponent.startTime);

        //Check if timer has elapsed, else continue to decrement each second.
        if (elapsedTime >= timerComponent.timerDuration) {
          indicator.setText('0s');
          timerComponent.timerComplete = true;
        } else {
          indicator.setText((timerComponent.timerDuration - elapsedTime).toString() + 's');
        }

        redrawComponents = true;
      } else {
        indicator.setText(timerComponent.timerDuration.toString() + 's');
      }
    }
  }

  //If the color changed, update the color and redraw the component.
  if (indicator.attrs.fill !== color1) {
    indicator.attrs.fill = color1;
    redrawComponents = true;
  }

  //If the opacity changed (due to a change in indication bit mask) update and redraw the component
  let opacity = indicator.getOpacity();
  if (mask && (opacity != OPACITY_KNOWN)) {
      indicator.opacity(OPACITY_KNOWN);
      redrawComponents = true;
  } else if ((! mask) && (opacity != OPACITY_UNKNOWN)) {
      indicator.opacity(OPACITY_UNKNOWN);
      redrawComponents = true;
  }

} // end UpdateIndication()

//***********************************************************************************************************************************************************************
// Components are LEDs, Pushbuttons, toggle switches, text, etc. that can be placed anywhere on the page
// to create a virtual panel that simulates the traditional hardware-based CLCPs.
function AddComponent(xpos, ypos, id_nbr) {
  let Component;
  let configureComponent;

  if (dropMode.quadrilateral) {
    Component = CreateQuadrilateral(xpos, ypos, id_nbr);
    configureComponent = true;
    if (!developerMode) {
      Component.listening(false);
    }
  }

  //create an arrow shape
  if (dropMode.arrow) {
    Component = CreateArrow(xpos, ypos, id_nbr);
    configureComponent = true;
    if (!developerMode) {
      Component.listening(false);
    }
  }

  //create an LED
  if (dropMode.LED) {
    Component = CreateLED(xpos, ypos, id_nbr);
    configureComponent = true;
    if (!developerMode) {
      Component.listening(false);
    }
  }

  //creat a round pushbutton
  if (dropMode.pb) {
    Component = CreatePushbutton(xpos, ypos, id_nbr);
    configureComponent = true;
  }

  //create a square pushbutton
  if (dropMode.pb_square) {
    Component = CreateSquarePushbutton(xpos, ypos, id_nbr);
    configureComponent = true;
  }

  //create a toggle switch
  if (dropMode.togSw || dropMode.togSwMom) {
    Component = CreateToggleSwitch(xpos, ypos, id_nbr);
    configureComponent = true;
  }

  //create a keyswitch
  if (dropMode.keyswitch) {
    Component = CreateKeySwitch(xpos, ypos, id_nbr);
    configureComponent = true;
  }

  // create a rotaryswitch
  if (dropMode.rotaryswitch) {
    Component = CreateRotarySwitch(xpos, ypos, id_nbr);
    configureComponent = true;
  }

  //create text
  if (dropMode.text) {
    Component = CreateText(xpos, ypos, id_nbr);
    configureComponent = true;

    Component.getChildren()[0].attrs.alignText();
    if (!developerMode) {
      Component.listening(false);
    }
  }

  //create timer text
  if (dropMode.timer) {
    Component = CreateTimer(xpos, ypos, id_nbr);
    configureComponent = true;
    if (!developerMode) {
      Component.listening(false);
    }
  }

  //If a new component was added, configure its properties and events
  if (configureComponent) {
    //initialize the indication properties for this component
    //An indication bit can be assigned for each color.
    for (let i=0; i < stdColors.length * 2; i++) {
      Component.attrs["indication" + i] = "0,0," + colorVals[i % stdColors.length];
    }

    //initialize the control properties for this component
    for (let i=0; i < Component.attrs.nbrCtrls; i++) {
      Component.attrs["control" + i] = "0";
    }

    //add the component to the layer
    component_layer.add(Component);
    //component_layer.draw();
  }

  //return component object
  return Component;
} // end AddComponent()

function CheckLockStatus(component) {
  //Check if passcode is defined and/or assigned for this control component
  //and then determine if it is locked.
  if (typeof component.attrs.passcode === 'undefined' ||
      component.attrs.passcode === null ||
      component.attrs.passcode === '') {
    //variable is undefined, null, or unassigned so unlock the control component
    component.attrs.unlocked = true;
    //console.log(selectedId + ' is unlocked');
  } else if (component.attrs.unlocked === false && initLatchedSwitches === false) {
    //if a passcode is defined for this component and the component is locked...

    //A valid passcode has not been entered or has timed out.
    //Request a passcode for this control.
    ShowPasscodeDialogBox(component.attrs.id);
    //console.log(selectedId + ' is locked');
  }
} // end CheckLockStatus()

function ShowComponentPosition(component) {
  document.getElementById("position_display_textbox").value = "  x = " + Math.round(component.getPosition().x) + ", y = " + Math.round(component.getPosition().y);
} // end ShowComponentPosition()

function ClearPositionDisplay() {
  document.getElementById("position_display_textbox").value = "";
  //console.log("Cleared position display");
} // end ClearPositionDisplay()

const hasColorInputSupport = () => {
  try {
      const input = document.createElement('input');
      input.type = "color";
      input.value = "!";
      return input.type === "color" && input.value !== "!";
  } catch (e) {
      return false;
  };
};

function ComponentProperties(componentObj) {
  // Build html string for component properties dialog box.
  // This string will set the innerHTML of the properties dialog box.
  if (designMode && selectedId !== "") {
    let propertiesStr = "";
    let toggleSw = false;

    // Create common properties options.
    propertiesStr =
    '<p class="prop-p">X Position:' +
    '<input id="x-pos-value" type="number" style="width: 208px; float: right" onchange="UpdateComponentPosition(\'' + componentObj.attrs.id + '\')"/></p>' +
    '<p class="prop-p">Y Position:' +
    '<input id="y-pos-value" type="number" style="width: 208px; float: right" onchange="UpdateComponentPosition(\'' + componentObj.attrs.id + '\')"/></p>' +
    '<p class="prop-p">Layer: (Max/Top:' + (getMaxZIndex()+1) + ')' +
    '<input id="z-value-front" type="button" value="Top" style="width: 70px; float: right" onclick="BringComponentToFront(\'' + componentObj.attrs.id + '\')"/>' +
    '<input id="z-value" type="number" min="1" step="1" style="width: 69px; float: right; text-align: center;" onchange="UpdateComponentZIndex(\'' + componentObj.attrs.id + '\')"/>' +
    '<input id="z-value-back" type="button" value="Bottom" style="width: 70px; float: right" onclick="SendComponentToBack(\'' + componentObj.attrs.id + '\')"/></p>';

    // Create PS_Square dimensions property options
    if (componentObj.attrs.type === "pb_square") {
      propertiesStr += '<p class="prop-p">Height:' +
      '<input id="squarepb-height" type="number" min="20" onchange="UpdateSquarePBDimensions(\'' + componentObj.attrs.id + '\')"/></p>' +
      '<p class="prop-p">Width:' +
      '<input id="squarepb-width" type="number" min="20" onchange="UpdateSquarePBDimensions(\'' + componentObj.attrs.id + '\')"/></p>';
    }

    // Create Quadrilateral point coordinate property options.
    if (componentObj.attrs.type === "quadrilateral") {
      propertiesStr += '<p class="prop-p">Point 1 (x,y):' +
      '<input id="point1y-value" type="number" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/>' +
      '<input id="point1x-value" type="number" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/></p>' +
      '<p class="prop-p">Point 2 (x,y):' +
      '<input id="point2y-value" type="number" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/>' +
      '<input id="point2x-value" type="number" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/></p>' +
      '<p class="prop-p">Point 3 (x,y):' +
      '<input id="point3y-value" type="number" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/>' +
      '<input id="point3x-value" type="number" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/></p>' +
      '<p class="prop-p">Point 4 (x,y):' +
      '<input id="point4y-value" type="number" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/>' +
      '<input id="point4x-value" type="number" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/></p>';
    }

    // Add the rotation property for quadrilateral or arrow component.
    if (componentObj.attrs.type === "quadrilateral" || componentObj.attrs.type === "arrow") {
      propertiesStr += '<p class="prop-p">Rotation:' +
      '<input id="rotation-value" type="number" style="width: 208px; float: right" onchange="RotateComponent(\'' + componentObj.attrs.id + '\')"/></p>';
    }

    if (componentObj.attrs.type !== "text" && componentObj.attrs.type !== "timer") {
      propertiesStr += '<p class="prop-p">Scale:' +
      '<input id="scale-value" type="number" style="width: 208px; float: right" onchange="UpdateScaleValue(\'' + componentObj.attrs.id + '\')"/></p>';
    }

    //if the component is not text, add line properties
    if (componentObj.attrs.type === "LED" || componentObj.attrs.type === "quadrilateral" || componentObj.attrs.type === "arrow") {
      propertiesStr += '<p class="prop-p">Line Width:' +
        '<input id="line-width-value" type="number" style="width: 208px; float: right" onchange="UpdateLineWidth(\'' + componentObj.attrs.id + '\')"/></p>';

      if (hasColorInputSupport()) {

        propertiesStr += '<p class="prop-p">Line Color:' +
        '<label class="transp-chkbx-lbl" for="line-transp-chkbx">Transparent</label>' +
        '<input type="checkbox" id="line-transp-chkbx" onchange="UpdateLineColor(\'' + componentObj.attrs.id + '\')"/>' +
        '<input type="color" class="color-input" id="line-color-input" onchange="UpdateLineColor(\'' + componentObj.attrs.id + '\')"/></p>';

        propertiesStr += '<p class="prop-p">Fill Color:' +
        '<label class="transp-chkbx-lbl" for="fill-transp-chkbx">Transparent</label>' +
        '<input type="checkbox" id="fill-transp-chkbx" onchange="UpdateDefaultFillColor(\'' + componentObj.attrs.id + '\')"/>' +
        '<input type="color" class="color-input" id="fill-color-input" onchange="UpdateDefaultFillColor(\'' + componentObj.attrs.id + '\')"/></p>';

      } else {

        propertiesStr += '<p class="prop-p">Line Color:' +
        '<select id="line-color-input" style="width: 210px; float: right" onchange="UpdateLineColor(\'' + componentObj.attrs.id + '\')">';
        // Create color choices
        for (let i = 0; i < stdColors.length; i++) {
          propertiesStr += '<option value="' + colorVals[i] + '">' + stdColors[i] + '</option>';
        }
        propertiesStr += '</select></p>';

        propertiesStr += '<p class="prop-p">Fill Color:' +
        '<select id="fill-color-input" style="width: 210px; float: right" onchange="UpdateDefaultFillColor(\'' + componentObj.attrs.id + '\')">';
        // Create color choices
        for (let i = 0; i < stdColors.length; i++) {
          propertiesStr += '<option value="' + colorVals[i] + '">' + stdColors[i] + '</option>';
        }
        propertiesStr += '</select></p>';
      }
    }

    if (componentObj.attrs.type === "text" || componentObj.attrs.type === "pb_square")
    {
      propertiesStr += '<p class="prop-p">Text:' +
      '<input id="text-value" type="text" style="width: 208px; float: right" onchange="UpdateTextValue(\'' + componentObj.attrs.id + '\')"/></p>';
    }

    if (componentObj.attrs.type === "text") {
      // Create textbox alignment options.
      propertiesStr += '<p class="prop-p">Text Box Alignment:' +
      '<select id="text-box-align" style="width: 210px; float: right" onchange="ChangeTextBoxAlignment(\'' + componentObj.attrs.id + '\')"></p>';
      propertiesStr += '<option>left</option><option>center</option><option>right</option></select>';
    }

    if (componentObj.attrs.type === "text" || componentObj.attrs.type === "timer" ||  componentObj.attrs.type === "pb_square") {
      propertiesStr += '<p class="prop-p">Text Size:' +
      '<select id="text-size" style="width: 210px; float: right" onchange="ChangeTextSize(\'' + componentObj.attrs.id + '\')">';

      // Create text size options.
      for (let i=10; i <= 44; i=i+2) {
        propertiesStr += '<option>' + i + '</option>';
      }
      propertiesStr += '</select></p>';

      if (componentObj.attrs.type === "text" || componentObj.attrs.type === "timer") {

        propertiesStr += '<p class="prop-p">' +
          'Rotation:<input id="rotation-value" type="number" style="width: 208px; float: right" onchange="RotateComponent(\'' + componentObj.attrs.id + '\')"/></p>';

        if (hasColorInputSupport()) {

          propertiesStr += '<p class="prop-p">Text Color:' +
          '<label class="transp-chkbx-lbl" for="text-transp-chkbx">Transparent</label>' +
          '<input type="checkbox" id="text-transp-chkbx" onchange="UpdateDefaultTextColor(\'' + componentObj.attrs.id + '\')"/>' +
          '<input type="color" class="color-input" id="text-color-input" onchange="UpdateDefaultTextColor(\'' + componentObj.attrs.id + '\')"/></p>';

        } else {

          propertiesStr += '<p class="prop-p">Text Color:' +
          '<select id="text-color-input" onchange="UpdateDefaultTextColor(\'' + componentObj.attrs.id + '\')">';
          // Create color choices
          for (let i = 0; i < stdColors.length; i++) {
            propertiesStr += '<option value="' + colorVals[i] + '">' + stdColors[i] + '</option>';
          }
          propertiesStr += '</select></p>';
        }
      }

      if (componentObj.attrs.type === "pb_square") {

        if (hasColorInputSupport()) {

          propertiesStr += '<p class="prop-p">Text Color:' +
          '<label class="transp-chkbx-lbl" for="text-transp-chkbx">Transparent</label>' +
          '<input type="checkbox" id="text-transp-chkbx" onchange="UpdateTextColor(\'' + componentObj.attrs.id + '\')"/>' +
          '<input type="color" class="color-input" id="text-color-input" onchange="UpdateTextColor(\'' + componentObj.attrs.id + '\')"/></p>';

        } else {

          propertiesStr += '<p class="prop-p">Text Color:' +
          '<select id="text-color-input" style="width: 210px; float: right" onchange="UpdateTextColor(\'' + componentObj.attrs.id + '\')">';
          // Create color choices
          for (let i = 0; i < stdColors.length; i++) {
            propertiesStr += '<option value="' + colorVals[i] + '">' + stdColors[i] + '</option>';
          }
          propertiesStr += '</select></p>';
        }
      }
    }

    if (componentObj.attrs.type === "timer") {
      propertiesStr += '<p class="prop-p">Timer Duration:' +
      '<input id="timer-duration-value" type="number" style="width: 208px; float: right" onchange="UpdateTimerDuration(\'' + componentObj.attrs.id + '\')"/></p>';
    }

    //If this is a toggle switch, show option for setting latching or momentary
    if (componentObj.attrs.type === "togSw" || componentObj.attrs.type === "togSwMom") {
      let momentary = "";
      let rotated = "";

      if (componentObj.attrs.type === "togSwMom") {
        momentary = "checked";
      }

      if (componentObj.attrs.rotation == 90) {
        rotated = "checked";
      }

      propertiesStr += '<p class="prop-p">Momentary (ON-OFF-ON):' +
      '<input type="checkbox"' + momentary + ' style="width: 20px; float: right" onchange="SetToggleSwitchType(\'' + componentObj.attrs.id + '\',this.checked)"/></p>' +
      '<p class="prop-p">Rotate 90 degrees:' +
      '<input type="checkbox"' + rotated + ' style="width: 20px; float: right" onchange="RotateToggleSwitch(\'' + componentObj.attrs.id + '\',this.checked)"/></p>';

      toggleSw = true;
    }

    //if has controls, add passcode property
    if (componentObj.attrs.nbrCtrls > 0) {
      propertiesStr += '<p class="prop-p">Passcode:' +
      '<input id="passcode-value" type="text" style="width: 208px; float: right" onchange="SetPasscode(\'' + componentObj.attrs.id + '\')"/></p>' +
      '<p class="prop-p">Passcode Timeout (s):' +
      '<input id="passcode-timeout-value" type="number" style="width: 208px; float: right" onchange="SetPasscodeTimeout(\'' + componentObj.attrs.id + '\')"/></p>';
    }

    // Loop through the number of controls for this component to create its control properties.
    for (let i=0; i < componentObj.attrs.nbrCtrls ; i++) {
      let labelNbr = i + 1;
      let descStr = "";

      if (toggleSw) {
        if (labelNbr == 1) {
          descStr = " (Up/Right Position)";
        }
        if (labelNbr == 2) {
          descStr = " (Down/Left Position)";
        }
      }

      if (componentObj.attrs.type === "keyswitch") {
        if (labelNbr == 1) {
          descStr = " (Up Position)";
        }
        if (labelNbr == 2) {
          descStr = " (Right Position)";
        }
        if (labelNbr == 3) {
          descStr = " (Left Position)";
        }
      }

      if (componentObj.attrs.type === "rotaryswitch") {
        switch(labelNbr) {
          case 1:
            descStr = " (12 o'clock)";
          break;
          case 2:
            descStr = " (2 o'clock)";
          break;
          case 3:
            descStr = " (4 o'clock)";
          break;
          case 4:
            descStr = " (6 o'clock)";
          break;
          case 5:
            descStr = " (8 o'clock)";
          break;
          case 6:
            descStr = " (10 o'clock)";
          break;
        }
      }

      //the last control is used for the unlock status
      if (labelNbr == componentObj.attrs.nbrCtrls) {
        descStr = " (Unlock Status)";
      }

      propertiesStr += '<p class="prop-p">Control Status ' + labelNbr + descStr + ': ' +
      '<select id="control' + i + '-lbl-opts" style="float: right; width: 210px" onchange="SetComponentControlProperty(\'' + componentObj.attrs.id + '\',\'control' + i + '\')">' +
      controlOptionsStr +
      '</select></p>';
    }

    if (componentObj.attrs.hasIndications) {
      //Add flash rate property
      propertiesStr += '<p class="prop-p">Flash Time (ms):' +
      '<input id="flash-time-value" type="number" style="width: 208px; float: right" onchange="UpdateComponentFlashTime(\'' + componentObj.attrs.id + '\')"/></p>';

      // -for now, loop through the indications.
      // -soon Add the assigned indications (and selectors to add/remove indication assignments).
      if (hasColorInputSupport()) {

        //Add the labels and drop down lists to select status names for the colored indication. Also add the lists to select the state option.
        //The state option determines whether the color is displayed when the bit state is true or when it is false.
        for (let indNbr=0; indNbr < stdColors.length * 2; indNbr++) {

          if (indNbr < stdColors.length) {
            propertiesStr += '<p class="prop-p">Indication:';
          } else {
            propertiesStr += '<p class="prop-p"> Flash Ind:';
          }
          propertiesStr += '<select id="indication'+ indNbr + '-lbl-opts" onchange="SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')">' +
          indicationOptionsStr +
          '</select>' +
          '<select id="indication'+ indNbr + '-val-opts" onchange="SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')">' +
          '<option>True</option><option>False</option>' +
          '</select>' +
          '<label class="transp-chkbx-lbl" for="indication' + indNbr + '-transp-chkbx">Transparent</label>' +
          '<input type="checkbox" id="indication' + indNbr + '-transp-chkbx" onchange="SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')"/>' +
          '<input type="color" id="indication' + indNbr + '-color-input" onchange="SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')"/></p>';
        }
      } else { // Color picker input not supported in browser, provide color dropdowns instead.
        //Add the labels and drop down lists to select status names for the colored indication. Also add the lists to select the state option.
        //The state option determines whether the color is displayed when the bit state is true or when it is false.
        let indNbr = 0;
        for (let i=0; i < stdColors.length * 2; i++) {

          if (i < stdColors.length) {
            propertiesStr += '<p class="prop-p">Indication:';
          } else {
          propertiesStr += '<p class="prop-p">Flash Ind:';
          }

          propertiesStr += '<select id="indication'+ indNbr + '-lbl-opts" onchange="SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')">' +
          indicationOptionsStr +
          '</select>' +
          '<select id="indication'+ indNbr + '-val-opts" onchange="SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')">' +
          '<option>True</option><option>False</option>' +
          '</select>' +
          '<select id="indication' + indNbr + '-color-input" style="width:110px; float: none;" "SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')">';
          // Create color choices
          for (let i = 0; i < stdColors.length; i++) {
            propertiesStr += '<option value="' + colorVals[i] + '">' + stdColors[i] + '</option>';
          }
          propertiesStr += '</select></p>';

          indNbr++;
        }
      }
    }

    // Build the final string and assign it to the properties dialog box.
    document.getElementById("properties_dialog").innerHTML = propertiesStr;

    // Now that the inner html for the properties is created,
    // get the property values for the current component.
    document.getElementById("x-pos-value").value = componentObj.attrs.x;
    document.getElementById("y-pos-value").value = componentObj.attrs.y;
    document.getElementById('z-value').value = componentObj.getZIndex() + 1;

    let colorVal = "";

    if (componentObj.attrs.type === "text" || componentObj.attrs.type === "pb_square") {
      document.getElementById("text-value").value = componentObj.find('Text')[0].attrs.text;
    }

    if (componentObj.attrs.type === "text") {
      document.getElementById("text-box-align").value = componentObj.find('Text')[0].attrs.align;
    }

    if (componentObj.attrs.type === "text" || componentObj.attrs.type === "timer") {
      let cmpAttr = componentObj.find('Text')[0].attrs;

      document.getElementById("text-size").value = cmpAttr.fontSize;

      if (hasColorInputSupport()) {
        colorVal = RGBAToHexA(cmpAttr.defaultColor);

        if (colorVal.substring(7) === "00") {
          document.getElementById("text-transp-chkbx").checked = true;
          document.getElementById("text-color-input").style.visibility = "hidden";
        }
        colorVal = colorVal.substring(0, 7);
      } else {
        colorVal = cmpAttr.defaultColor;
      }

      document.getElementById("text-color-input").value = colorVal;

    } else {
      document.getElementById("scale-value").value = Math.round(componentObj.attrs.scaleX * 100);
    }

    if (componentObj.attrs.type === "pb_square") {
      let cmpAttr = componentObj.find('Text')[0].attrs;
      document.getElementById("text-size").value = cmpAttr.fontSize;

      if (hasColorInputSupport()) {
        colorVal = RGBAToHexA(cmpAttr.fill);

        if (colorVal.substring(7) === "00") {
          document.getElementById("text-transp-chkbx").checked = true;
          document.getElementById("text-color-input").style.visibility = "hidden";
        }
        colorVal = colorVal.substring(0, 7);
      } else {
        colorVal = cmpAttr.fill;
      }
      document.getElementById("text-color-input").value = colorVal;

      // Now get the current dimensions.
      cmpAttr = componentObj.find('Rect')[0];
      document.getElementById("squarepb-height").value = cmpAttr.height();
      document.getElementById("squarepb-width").value = cmpAttr.width();

    }

    if (componentObj.attrs.type === "quadrilateral") {

      let quad = componentObj.find('.indicator')[0].attrs;
      document.getElementById("point1x-value").value = quad.point1x;
      document.getElementById("point1y-value").value = quad.point1y;
      document.getElementById("point2x-value").value = quad.point2x;
      document.getElementById("point2y-value").value = quad.point2y;
      document.getElementById("point3x-value").value = quad.point3x;
      document.getElementById("point3y-value").value = quad.point3y;
      document.getElementById("point4x-value").value = quad.point4x;
      document.getElementById("point4y-value").value = quad.point4y;
    }

    if (componentObj.attrs.type === "timer") {
      document.getElementById("timer-duration-value").value = componentObj.find('Text')[0].attrs.timerDuration;
    }

    if (componentObj.attrs.type === "text" || componentObj.attrs.type === "quadrilateral" ||
      componentObj.attrs.type === "arrow" || componentObj.attrs.type === "timer") {

      document.getElementById("rotation-value").value = componentObj.attrs.rotation;
    }

    if (componentObj.attrs.type === "LED" || componentObj.attrs.type === "quadrilateral" || componentObj.attrs.type === "arrow") {
      let cmpAttr = componentObj.find('.indicator')[0].attrs;

      document.getElementById("line-width-value").value = cmpAttr.strokeWidth;

      if (hasColorInputSupport()) {
        colorVal = RGBAToHexA(cmpAttr.stroke);

        if (colorVal.substring(7) === "00") {
          document.getElementById("line-transp-chkbx").checked = true;
          document.getElementById("line-color-input").style.visibility = "hidden";
        }
        colorVal = colorVal.substring(0,7);
      }  else {
        colorVal = cmpAttr.stroke;
      }

      document.getElementById("line-color-input").value = colorVal;

      if (hasColorInputSupport()) {
        colorVal = RGBAToHexA(cmpAttr.defaultColor);

        if (colorVal.substring(7) === "00") {
          document.getElementById("fill-transp-chkbx").checked = true;
          document.getElementById("fill-color-input").style.visibility = "hidden";
        }
        colorVal = colorVal.substring(0,7);
      } else {
        colorVal = cmpAttr.defaultColor;
      }

      document.getElementById("fill-color-input").value = colorVal;
    }

    if (componentObj.attrs.nbrCtrls > 0) {
      document.getElementById("passcode-value").value = componentObj.attrs.passcode;
      document.getElementById("passcode-timeout-value").value = componentObj.attrs.timeout;
    }

    for (let i=0; i < componentObj.attrs.nbrCtrls; i++) {
      document.getElementById("control" + i + "-lbl-opts").selectedIndex = componentObj.attrs["control" + i];
    }

    if (componentObj.attrs.hasIndications) {
      document.getElementById("flash-time-value").value = componentObj.attrs.flashTime;

      for (let indIdx = 0; indIdx < stdColors.length * 2; indIdx++) {

        let indObjAttrs = componentObj.attrs["indication" + indIdx].split(",");
        let state = document.getElementById("indication" + indIdx + "-val-opts");
        let color = document.getElementById("indication" + indIdx + "-color-input");
        let transp = document.getElementById("indication" + indIdx + "-transp-chkbx");

        let labelIdx = indObjAttrs.shift();

        document.getElementById("indication" + indIdx + "-lbl-opts").selectedIndex = labelIdx;
        state.selectedIndex = indObjAttrs.shift();

        let colorVal = indObjAttrs.join();
        let vis = (labelIdx == 0) ? "hidden" : "visible";

        state.style.visibility = vis;
        color.style.visibility = vis;

        if (transp == null) {
          color.value = colorVal;
        } else {

          transp.style.visibility = vis;
          transp.labels[0].style.visibility = vis;

          colorVal = RGBAToHexA(colorVal);

          if (colorVal.substring(7) === "00") {
            transp.checked = true;
            color.style.visibility = "hidden";
          }
          colorVal = colorVal.substring(0,7);
        }

        color.value = colorVal;
      }
    }

  } // end if (designMode && selectedId !== "")
} // end ComponentProperties()

function ShowPropertiesDialog() {
  $( "#properties_dialog" ).dialog( "option", "title", "Properties" );

  //Open the properties dialog box
  $( "#properties_dialog" ).dialog( "open" );

  //Request the control labels from the QLCP GUI Developer tool and create the control status name options.
  $.get("http://localhost:"+serverPort+"/control_labels/", function(data){

    UpdateControlLabels(data);

    // Request the indication labels from the QLCP GUI Developer tool and create the indication status name options.
    $.get("http://localhost:"+serverPort+"/indication_labels/", function(data){

      UpdateIndicationLabels(data);

      }, "text");
    }, "text");

  //console.log("Selected ID = " + selectedId);

  if (selectedId !== "") {
    ComponentProperties(stage.find('#'+ selectedId)[0]);
  }
} // end ShowPropertiesDialog()

function ClearPreviousSelection()
{
  //Deselect the previous component
  if (selectedId !== "")
  {
    let prevObj = stage.find('#'+ selectedId)[0];
    prevObj.attrs.selected = false;
    prevObj.find('.selection')[0].attrs.hide();
    component_layer.draw();
  }
  selectedId = "";

  ClearPositionDisplay();

  //clear properties dialog box
  document.getElementById("properties_dialog").innerHTML = "";
} // end ClearPreviousSelection()

//this function shows the selected component and clears the selection
//of the previously selected component. It also adds or removes a selected
//component from a group selection. This function depends on requestAnimationFrame
//to call UpdateComponentLayer in order to see the changes.
function ShowSelectedComponent(componentObj) {
  ClearPreviousSelection();

  if (typeof componentObj !== 'undefined') {
    selectedId = componentObj.attrs.id;
    ShowComponentPosition(componentObj);

    //show the component as selected
    componentObj.find('.selection')[0].attrs.show();

    if ($( "#properties_dialog" ).dialog( "isOpen" )) {
      ComponentProperties(componentObj);
    }

    //If there is a group of selected components...
    if (selectedComponents.length > 0) {
      //If the Ctrl key is down...
      if (ctrlKeyDown) {
        //add the selected component to the selected group if not already selected
        if (!componentObj.attrs.selected) {
          console.log("add to selection");
          componentObj.attrs.selected = true;
          selectedComponents.push(componentObj);
        } else {
          //remove the component from the group
          let index = selectedComponents.indexOf(componentObj);
          console.log("remove index " + index + " from selection");
          if (index > -1) {
            selectedComponents.splice(index, 1);
          }
          componentObj.attrs.selected = false;
          ClearPreviousSelection();
        }
        compInSelectedGroup = true;
      }
    }
  }

  // console.log("Selected id = " + selectedId);
} // end ShowSelectedComponent()

function SetToggleSwitchType(id, checked) {
  let componentObj = stage.find('#'+id)[0];

  if (checked) {
    componentObj.attrs.type = "togSwMom";
    componentObj.find('.up')[0].visible(false);
    componentObj.find('.down')[0].visible(false);
    componentObj.find('.center')[0].visible(true);
  } else {
    componentObj.attrs.type = "togSw";
    componentObj.find('.up')[0].visible(false);
    componentObj.find('.down')[0].visible(true);
    componentObj.find('.center')[0].visible(false);
  }

  component_layer.draw();
} // end SetToggleSwitchType()

function SetComponentControlProperty(id, property) {
  //This function is called when control status name (bit assignement) is changed to update the saved values.
  let statusListId = property + "-lbl-opts";
  let componentObj = stage.find('#'+id)[0];
  componentObj.attrs[property] = document.getElementById(statusListId).selectedIndex;
} // end SetComponentControlProperty()

function SetComponentIndicationProperty(id, property) {
  // This function is called when an indication color, status name, or state option is changed to update the saved values.
  // The state option determines whether the color is displayed when the bit state is true or when it is false.
  let textObj = stage.find('#'+id)[0];
  let state = document.getElementById(property + "-val-opts");
  let color = document.getElementById(property + "-color-input");
  let transp = document.getElementById(property + "-transp-chkbx");

  let labelIdx = document.getElementById(property + "-lbl-opts").selectedIndex;
  let stateIdx = state.selectedIndex;

  // Update the visibility based on having an assigned label or not.
  let vis = (labelIdx == 0) ? "hidden" : "visible";

  state.style.visibility = vis;
  color.style.visibility = vis;

  if (transp == null) {
    color = color.value;
  } else {

    transp.style.visibility = vis;
    transp.labels[0].style.visibility = vis;

    if (transp.checked) {
      color.style.visibility = "hidden";
      transp = '00';
    } else {
      transp = 'FF';
    }
    color = hexAToRGBA(color.value + transp);
  }

  textObj.attrs[property] = labelIdx + "," + stateIdx + "," + color;
} // end SetComponentIndicationProperty()

function UpdateTextValue(id) {
  let textObj = stage.find('#'+id)[0].find('Text')[0];
  textObj.setText(document.getElementById("text-value").value);

  if (textObj.getParent().attrs.type === 'pb_square' || textObj.getParent().attrs.type === 'text') {
    textObj.attrs.alignText();
  }

  //update the selection border to account for line width
  stage.find('#'+id)[0].find('.selection')[0].attrs.update();

  component_layer.draw();
} // end UpdateTextValue()

function UpdateDefaultTextColor(id) {
  const component = stage.find('#'+id)[0].find('Text')[0];
  let color = document.getElementById("text-color-input");
  let transp = document.getElementById("text-transp-chkbx");

  if (transp != null ) {
    if (transp.checked) {
      color.style.visibility = "hidden";
      transp = '00';
    } else {
      transp = 'FF';
      color.style.visibility = "visible";
    }
    color = hexAToRGBA(color.value + transp);
  } else {
    color = color.value;
  }

  component.attrs.defaultColor = color;
  component_layer.draw();
} // end UpdateDefaultTextColor()

function UpdateTextColor(id) {
  const component = stage.find('#'+id)[0].find('Text')[0];
  let color = document.getElementById("text-color-input");
  let transp = document.getElementById("text-transp-chkbx");

  if (transp != null ) {
    if (transp.checked) {
      color.style.visibility = "hidden";
      transp = "00";
    } else {
      transp = "FF";
      color.style.visibility = "visible";
    }
    color= hexAToRGBA(color.value + transp);
  } else {
    color = color.value;
  }

  component.attrs.fill = color;
  component_layer.draw();
} // end UpdateTextColor()

function UpdateLineColor(id) {
  const component = stage.find('#'+id)[0].find('.indicator')[0];
  let color = document.getElementById("line-color-input");
  let transp = document.getElementById("line-transp-chkbx");

  if (transp != null ) {
    if (transp.checked) {
      color.style.visibility = "hidden";
      transp = "00";
    } else {
      transp = "FF";
      color.style.visibility = "visible";
    }
    color = hexAToRGBA(color.value + transp);
  } else {
    color = color.value;
  }

  component.stroke(color);
  component_layer.draw();
} // end UpdateLineColor()

function UpdateDefaultFillColor(id) {
  const component = stage.find('#'+id)[0].find('.indicator')[0];
  let color = document.getElementById("fill-color-input");
  let transp = document.getElementById("fill-transp-chkbx")

  if (transp != null ) {
    if (transp.checked) {
      color.style.visibility = "hidden";
      transp = "00";
    } else {
      transp = "FF";
      color.style.visibility = "visible";
    }
    color = hexAToRGBA(color.value + transp);
  } else {
    color = color.value;
  }

  component.attrs.defaultColor = color;
  component_layer.draw();
 }// end UpdateDefaultFillColor()

function ChangeTextBoxAlignment(id) {
  let textObj = stage.find('#'+id)[0].find('Text')[0];
  textObj.align(document.getElementById("text-box-align").value);
  lastTextAlign = textObj.align();
  textObj.attrs.alignText();
} // end ChangeTextBoxAlignment()

function ChangeTextSize(id) {
  let textObj = stage.find('#'+id)[0].find('Text')[0];
  textObj.fontSize(document.getElementById("text-size").value);

  //refresh the text value so that selection border resizing will work
  textObj.setText(textObj.attrs.text);

  if (textObj.getParent().attrs.type === 'pb_square' ||
      textObj.getParent().attrs.type === 'text') {
    textObj.attrs.alignText();
  }

  //update the selection border to account for line width
  stage.find('#'+id)[0].find('.selection')[0].attrs.update();

  component_layer.draw();
} // end ChangeTextSize()

function IncreaseBackgroundImageSize() {
  //if there is an image, resize it based on the current scale factor
  if (bgImage_layer.hasChildren()) {
    let bgImg = bgImage_layer.getChildren()[0];
    let sz = bgImg.size();
    bgImg.size({
      width: sz.width + 10,
      height: sz.height + (10 * sz.height/sz.width)
    });
  }
  bgImage_layer.draw();
} // end IncreaseBackgroundImageSize()

function DecreaseBackgroundImageSize() {
  //if there is an image, resize it based on the current scale factor
  if (bgImage_layer.hasChildren()) {
    let bgImg = bgImage_layer.getChildren()[0];
    let sz = bgImg.size();
    bgImg.size({
      width: sz.width - 10,
      height: sz.height - (10 * sz.height/sz.width)
    });
  }
  bgImage_layer.draw();
} // end DecreaseBackgroundImageSize()

function RemoveBackgroundImage() {
  if (bgImage_layer.hasChildren()) {
    bgImage_layer.removeChildren();
    bgImage_layer.clear();
    bgImage_layer.draw();
  }
} // end RemoveBackgroundImage()

function TogglePositionLock() {

  const allObjects = component_layer.getChildren();

  if (TogglePositionLock.positionLocked == true) {
    TogglePositionLock.positionLocked = false;
    document.getElementById('lockBtnImg').src="images/unlocked-26.png"
    document.getElementById('lock_positions_button').title ="Lock Component Positions";
    allObjects.draggable(true);
  } else {
    TogglePositionLock.positionLocked = true;
    document.getElementById('lockBtnImg').src="images/locked-26.png"
    document.getElementById('lock_positions_button').title ="Unlock Component Positions";

    allObjects.draggable(false);
  }
} // end TogglePositionLock()

function UpdateComponentPosition(id)
{
  let component = stage.find('#'+id)[0];

  component.position({
    x: Math.floor(Number(document.getElementById("x-pos-value").value)),
    y: Math.floor(Number(document.getElementById("y-pos-value").value))
  });
  component_layer.draw();
  //console.log("x = " + component.attrs.x + ", y= " + component.attrs.y);
  ShowComponentPosition(component);
} // end UpdateComponentPosition()

function getMaxZIndex() {
  //find the maximum zIndex.
  let maxZ = 0;
  for (let j=0; j < component_layer.children.length; j++) {
    let cmpObj = component_layer.children[j];
    if (cmpObj.getZIndex() > maxZ) {
      maxZ = cmpObj.getZIndex();
    }
  }
  return maxZ;
} // end getMaxZIndex()

function UpdateComponentZIndex(id) {
  let comp = stage.find('#'+id)[0];
  let z = Math.round(document.getElementById("z-value").value)-1;
  let maxZ = getMaxZIndex();

  if (z < 0) {
    z = 0;
  } else if (z > maxZ) {
    z = maxZ;
  }

  comp.setZIndex(z);
  document.getElementById("z-value").value = z+1;

  component_layer.draw();
} // end UpdateComponentZIndex()

function BringComponentToFront(id) {
  let component = stage.find('#'+id)[0];
  component.setZIndex(getMaxZIndex());
  document.getElementById("z-value").value = component.getZIndex()+1;
  component_layer.draw();
} // end BringComponentToFront()

function SendComponentToBack(id) {
  let component = stage.find('#'+id)[0];
  component.setZIndex(0);
  document.getElementById("z-value").value = 1;
  component_layer.draw();
} // end SendComponentToBack()

function UpdateLineWidth(id) {
  if (Number.isInteger(Number(document.getElementById("line-width-value").value))) {
    let component = stage.find('#'+id)[0].find('.indicator')[0];

    if (document.getElementById("line-width-value").value <= 0) {
      document.getElementById("line-width-value").value = 0;
      component.strokeEnabled(false);
    } else {
      component.strokeEnabled(true);
    }

    if (document.getElementById("line-width-value").value > 20) {
      document.getElementById("line-width-value").value = 20;
    }

    component.strokeWidth(document.getElementById("line-width-value").value);

    //update the selection border to account for line width
    stage.find('#'+id)[0].find('.selection')[0].attrs.update();

    component_layer.draw();
  }
} // end UpdateLneWidth()

function UpdateScaleValue(id) {
  let component = stage.find('#'+id)[0];

  component.scale({
    x: document.getElementById("scale-value").value/100,
    y: document.getElementById("scale-value").value/100
  });

  //update the selection border to account for line width
  stage.find('#'+id)[0].find('.selection')[0].attrs.update();

  component_layer.draw();
} // end UpdateScaleIndex()

function UpdateSquarePBDimensions(id) {

  let pbComps = stage.find('#'+id)[0].find('Rect');
  let hght = Math.round(document.getElementById("squarepb-height").value);
  let wdth = Math.round(document.getElementById("squarepb-width").value);

  if (hght < 20) {
    hght = 20;
  }

  if (wdth < 20) {
    wdth = 20;
  }

  // Update the input field.
  document.getElementById("squarepb-height").value = hght;
  document.getElementById("squarepb-width").value = wdth;

  // Update the switch base dimensions.
  pbComps[0].height(hght);
  pbComps[0].width(wdth);

  // Update the indicator dimensions.
  pbComps[1].width(pbComps[0].width()-10);
  pbComps[1].height(pbComps[0].height()-10);

  // Update the selection border.
  pbComps[2].width(pbComps[0].width());
  pbComps[2].height(pbComps[0].height());

  // Re-align the text.
  UpdateTextValue(id);
} // end UpdateSquarePBDimensions()

function UpdateQuadrilateralCoordinates(id) {
  let minPixelDiff = 2;

  //get the quad shape from the stage
  let quad = stage.find('#'+id)[0].find('.indicator')[0].attrs;

  //check coordinates...
  //first, round any decimal values
  document.getElementById("point1x-value").value = Math.round(document.getElementById("point1x-value").value);
  document.getElementById("point2x-value").value = Math.round(document.getElementById("point2x-value").value);
  document.getElementById("point3x-value").value = Math.round(document.getElementById("point3x-value").value);
  document.getElementById("point4x-value").value = Math.round(document.getElementById("point4x-value").value);
  document.getElementById("point1y-value").value = Math.round(document.getElementById("point1y-value").value);
  document.getElementById("point2y-value").value = Math.round(document.getElementById("point2y-value").value);
  document.getElementById("point3y-value").value = Math.round(document.getElementById("point3y-value").value);
  document.getElementById("point4y-value").value = Math.round(document.getElementById("point4y-value").value);

  //2x cannot be less than minPixelDiff greater than 1x
  if (Number(document.getElementById("point2x-value").value) - Number(document.getElementById("point1x-value").value) < minPixelDiff) {
    document.getElementById("point2x-value").value = Number(document.getElementById("point1x-value").value) + minPixelDiff;
  }

  //3x cannot be less than minPixelDiff greater than 4x
  if (Number(document.getElementById("point3x-value").value) - Number(document.getElementById("point4x-value").value) < minPixelDiff) {
    document.getElementById("point3x-value").value = Number(document.getElementById("point4x-value").value) + minPixelDiff;
  }

  //4y cannot be less than minPixelDiff greater than 1y
  if (Number(document.getElementById("point4y-value").value) - Number(document.getElementById("point1y-value").value) < minPixelDiff) {
    document.getElementById("point4y-value").value = Number(document.getElementById("point1y-value").value) + minPixelDiff;
  }

  //3y cannot be less than minPixelDiff greater than 2y
  if (Number(document.getElementById("point3y-value").value) - Number(document.getElementById("point2y-value").value) < minPixelDiff) {
    document.getElementById("point3y-value").value = Number(document.getElementById("point2y-value").value) + minPixelDiff;
  }

  quad.point1x = Number(document.getElementById("point1x-value").value);
  quad.point1y = Number(document.getElementById("point1y-value").value);
  quad.point2x = Number(document.getElementById("point2x-value").value);
  quad.point2y = Number(document.getElementById("point2y-value").value);
  quad.point3x = Number(document.getElementById("point3x-value").value);
  quad.point3y = Number(document.getElementById("point3y-value").value);
  quad.point4x = Number(document.getElementById("point4x-value").value);
  quad.point4y = Number(document.getElementById("point4y-value").value);

  //update the selection border since the size has changed
  stage.find('#'+id)[0].find('.selection')[0].attrs.update();

  component_layer.draw();
} // end UpdateQuadrilateralCoordinates()

function RotateComponent(id) {
  let component = stage.find('#'+id)[0];

  if ((document.getElementById("rotation-value").value < 0)||(document.getElementById("rotation-value").value >= 360)) {
    document.getElementById("rotation-value").value = 0;
  }

  component.rotation(document.getElementById("rotation-value").value);
  component_layer.draw();
} // end RotateComponent()

function RotateToggleSwitch(id, checked) {
  let component = stage.find('#'+id)[0];

  if (checked) {
    component.rotation(90);
  } else {
    component.rotation(0);
  }

  component_layer.draw();
} // end RotateToggleSwitch()

function UpdateComponentFlashTime(id) {
  let component = stage.find('#'+id)[0];

  if (document.getElementById("flash-time-value").value < 0) {
    document.getElementById("flash-time-value").value = 0;
  }

  if (document.getElementById("flash-time-value").value > 65535) {
    document.getElementById("flash-time-value").value = 65535;
  }

  component.attrs.flashTime = document.getElementById("flash-time-value").value;
} // end UpdateComponentFlashTime()

function UpdateTimerDuration(id) {
  let component = stage.find('#'+id)[0].find('.indicator')[0];

  if (document.getElementById("timer-duration-value").value < 0) {
    document.getElementById("timer-duration-value").value = 0;
  }

  if (document.getElementById("timer-duration-value").value > 65535) {
    document.getElementById("timer-duration-value").value = 65535;
  }

  component.attrs.timerDuration = document.getElementById("timer-duration-value").value;
  component.setText(component.attrs.timerDuration.toString() + 's');

  //update the selection border to account for line width
  stage.find('#'+id)[0].find('.selection')[0].attrs.update();
} // end UpdateTimerDuration()

function Duplicate() {
  let duplicateComponents = new Array();

  DeselectDropComponentButtons();

  //Check for single or multiple-component selection
  if (selectedComponents.length === 0) {
    if (selectedId !== "") {
      let cmp = stage.find('#'+ selectedId)[0];
      if (cmp != null) {
        DuplicateComponent(cmp, false);
      }
    }
  }
  else {//multiple components are selected
    for (let i = 0; i < selectedComponents.length; i++) {
      let duplicate = DuplicateComponent(selectedComponents[i], true);

      //save duplicate component
      duplicateComponents.push(duplicate);
    }

    //clear original group of selected components
    ClearSelectedComponents();

    //Make the duplicate components the currently selected components.
    selectedComponents = duplicateComponents;
  }
} // end Duplicate()

function DuplicateComponent(component, groupSelected) {
  //This array contains the key/value pairs of attributes of copied component
  let attributes = new Array();
  let cmpType = component.attrs.type;

  //console.log("Duplicate component type " + cmpType);
  if (!groupSelected) {
    ClearPreviousSelection();
  }
  dropMode[cmpType] = true;
  let duplicate = AddComponent(component.attrs.x, component.attrs.y, ++idNbr);

  //copy all attributes for the component
  for (let key in component.attrs) {
    duplicate.attrs[key] = component.attrs[key]; //this makes sure controls and indication bits are copied
    attributes[key] = component.attrs[key]; //create attributes array to use with SetComponentProperties function.
    //console.log("Attribute " + key + "=" + component.attrs[key]);
  }

  //copy all the attributes for the component's indicator object
  let obj = component.find('.indicator')[0];
  if (typeof obj != 'undefined') {
    for (let key in obj.attrs) {
      attributes[key] = component.find('.indicator')[0].attrs[key];
    }
  }
  obj = component.find('Text')[0]
  //copy all the attributes for the component's text object
  if (typeof obj !== 'undefined') {
    for (let key in obj.attrs) {
      attributes[key] = obj.attrs[key];
    }
  }

  //offset the location of the duplicate component
  duplicate.attrs.x += 50;
  //duplicate.attrs.y += 50;

  //set the new id
  duplicate.attrs.id = 'c' + idNbr;
  selectedId = duplicate.attrs.id;

  SetComponentProperties(duplicate, attributes);
  //console.log("Duplicated component " + component.attrs.id + "; New component id = " +
  //            duplicate.attrs.id + "; Selected id = " + selectedId);

  if (groupSelected === false) {
    ShowSelectedComponent(duplicate);
    ComponentProperties(duplicate);
  } else {
    duplicate.find('.selection')[0].attrs.show();
  }

  //update the selection border
  if (typeof duplicate.find('Text')[0] !== 'undefined'){
    duplicate.find('Text')[0].setText(duplicate.find('Text')[0].attrs.text);
  }
  duplicate.find('.selection')[0].attrs.update();

  dropMode[cmpType] = false;

  return duplicate;
} // end DuplicateComponent()

function ShowPasscodeDialogBox(id) {
  //Build the html string for the code buttons of the passcode dialog box.
  document.getElementById("passcode_dialog").innerHTML = '<p><input id="btn_1" type="button" value="1" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(1)"/>'
    + '<input id="btn_2" type="button" value="2" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(2)"/>'
    + '<input id="btn_3" type="button" value="3" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(3)"/></p>'
    + '<p><input id="btn_4" type="button" value="4" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(4)"/>'
    + '<input id="btn_5" type="button" value="5" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(5)"/>'
    + '<input id="btn_6" type="button" value="6" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(6)"/></p>'
    + '<p><input id="btn_7" type="button" value="7" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(7)"/>'
    + '<input id="btn_8" type="button" value="8" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(8)"/>'
    + '<input id="btn_9" type="button" value="9" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(9)"/></p>'
    + '<p><input id="btn_enter" type="button" value="Enter" style="font-size: 20px; height: 100px; width: 100px; float: left" onclick="PasscodeEntered(\'' + id + '\')"/>'
    + '<input id="btn_0" type="button" value="0" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(0)"/>'
    + '<input id="btn_cancel" type="button" value="Cancel" style="font-size: 20px; height: 100px; width: 100px; float: left" onclick="ClosePasscodeDialog()"/></p>';

  $( "#passcode_dialog" ).dialog( "option", "title", "Enter Passcode" );

  //Open the passcode dialog box
  $( "#passcode_dialog" ).dialog( "open" );

  // Remove focus on all buttons within the div with class ui-dialog
  $('.ui-dialog :button').blur();
} // end ShowPasscodeDialogBox()

function PasscodeDigitEntered(digit) {
  passcode += digit;
} // end PasscodeDigitEntered()

function ClosePasscodeDialog() {
  passcode = "";
  $( "#passcode_dialog" ).dialog( "close" );
} // end ClosePasscodeDialog()

function PasscodeEntered(id) {
  //check passcode
  let component = stage.find('#'+id)[0];

  if (component.attrs.passcode === passcode) {
    component.attrs.unlocked = true;

    //send the control bit for unlock status
    let n = component.attrs.nbrCtrls - 1;
    SendControl(component.attrs["control" + n]-1, 1);

    $( "#message_dialog" ).dialog( "option", "title", "Passcode Accepted" );

    document.getElementById("message_dialog").innerHTML = '<div><img style="float: left; margin-top: 5px; width: 36px" src="images/info32px.png"/>'
      + '<div style="float: left; margin-left: 15px; margin-top: 6px; width: 350px">This control is now unlocked.<br/><br/></div></div>';

    ClosePasscodeDialog();

    //set timeout for passcode
    setTimeout(PasscodeTimeout, component.attrs.timeout * 1000, id);

    //console.log(Date() + ' Component ' + id + ' passcode accepted; control unlocked');
  } else {
    component.attrs.unlocked = false;

    $( "#message_dialog" ).dialog( "option", "title", "Incorrect Passcode" );

    document.getElementById("message_dialog").innerHTML = '<div><img style="float: left; margin-top: 5px; width: 36px" src="images/warning_icon.png"/>'
      + '<div style="float: left; margin-left: 15px; margin-top: 6px; width: 350px">This control will remain locked.<br/><br/></div></div>';
  }

  ShowMessageDialog();

  passcode = "";
} // end PasscodeEntered()

function SetPasscode(id) {
  let component = stage.find('#'+id)[0];
  component.attrs.passcode = document.getElementById("passcode-value").value;
} // end SetPasscode()

function SetPasscodeTimeout(id) {
  let component = stage.find('#'+id)[0];

  if (document.getElementById("passcode-timeout-value").value < 0) {
    document.getElementById("passcode-timeout-value").value = 0;
  }

  if (document.getElementById("passcode-timeout-value").value > 65535) {
    document.getElementById("passcode-timeout-value").value = 65535;
  }

  component.attrs.timeout = document.getElementById("passcode-timeout-value").value;
} // end SetPasscodeTimeout()

function PasscodeTimeout(id) {
  let component = stage.find('#'+id)[0];
  let elapsed = SecondsElapsed(lastClickTime);

  //Check time elapsed since last control was sent
  if (elapsed > component.attrs.timeout) {
    //clear the passcode result to force re-entry
    component.attrs.unlocked = false;

    //send the control bit for unlock status
    let n = component.attrs.nbrCtrls - 1;
    SendControl(component.attrs["control" + n]-1, 0);

    //console.log(Date() + ' Component ' + id + ' passcode timeout; control locked');
  } else {
    //start the timeout timer again, subtracting the time since last click;
    //otherwise, the timeout will be this time difference plus the timeout value (too long).
    let adjustedTimeout = (component.attrs.timeout - elapsed)*1000;
    setTimeout(PasscodeTimeout, adjustedTimeout, id);
  }
} // end PasscodeTimeout()

//When selecting multiple components, this function draws the selection box
//This function depends on requestAnimationFrame to redraw the component layer.
function CreateSelectionBox() {
  if (draggingToSelectComponents) {
    let mousepos = stage.getPointerPosition();
    let x = Math.round(mousepos.x);
    let y = Math.round(mousepos.y);

    if (!selectionBoxCreated) {
      initX = x;
      initY = y;

      //create the selection box
      selectionBox = new Konva.Rect({
        x: initX,
        y: initY,
        height: 1,
        width: 1,
        fill: "rgba(0,0,0,0)",
        stroke: "rgba(255,255,255,1)",
        strokeWidth: 1
      });
      component_layer.add(selectionBox);
      selectionBoxCreated = true;
    } else {
      let height;
      let width;
      let newX;
      let newY;

      if (x > initX)
        newX = initX;
      else
        newX = x;

      if (y > initY)
        newY = initY;
      else
        newY = y;

      height = Math.abs(Math.abs(y) - Math.abs(initY));
      width = Math.abs(Math.abs(x) - Math.abs(initX));

      selectionBox.setHeight(height);
      selectionBox.setWidth(width);
      selectionBox.setX(newX);
      selectionBox.setY(newY);
    }
  }
} // end CreateSelectionBox()

//This function finds the components in the selected area and highlights them.
function ShowSelectedGroup() {
  selectedComponents.length = 0;

  initX = 0;
  initY = 0;

  let arGroups = component_layer.getChildren();

  for (let i = 0; i < arGroups.length; i++) {
    let grp = arGroups[i];

    if (typeof grp.attrs.id !== 'undefined') {
      let pos = selectionBox.getAbsolutePosition();

      //get the extents of the selection box
      let selRecXStart = parseInt(pos.x);
      let selRecXEnd = parseInt(pos.x) + parseInt(selectionBox.attrs.width);
      let selRecYStart = parseInt(pos.y);
      let selRecYEnd = parseInt(pos.y) + parseInt(selectionBox.attrs.height);

      //get the extents of the group to compare to
      let grpXStart = parseInt(grp.attrs.x);
      let grpXEnd = parseInt(grp.attrs.x) + parseInt(grp.width());
      let grpYStart = parseInt(grp.attrs.y);
      let grpYEnd = parseInt(grp.attrs.y) + parseInt(grp.height());

      //Are we inside the selction area?
      if ((selRecXStart <= grpXStart && selRecXEnd >= grpXEnd) &&
          (selRecYStart <= grpYStart && selRecYEnd >= grpYEnd)) {
        grp.find('.selection')[0].attrs.show();
        grp.attrs.selected = true;
        selectedComponents.push(grp);
      }
    }
  }
} //  end ShowSelectedGroup()

//When multiple components are selected, this function clears the selection highlights.
//This function depends on requestAnimationFrame to redraw the component layer.
function ClearSelectedComponents() {
  //for each selected component
  for (let i=0; i < selectedComponents.length; i++) {
    selectedComponents[i].find('.selection')[0].attrs.hide();
    selectedComponents[i].attrs.selected = false;
  }

  //clear the array of selected components
  selectedComponents.length = 0;
  redrawComponents = true;
} // end ClearSelectedComponents()

function ToggleFullScreen() {
  if ((document.fullscreenElement !== undefined && document.fullscreenElement === null) ||
      (document.msFullscreenElement !== undefined && document.msFullscreenElement === null) ||
      (document.mozFullScreenElement !== undefined && !document.mozFullScreenElement) ||
      (document.webkitFullscreenElement !== undefined && !document.webkitFullscreenElement)) {
    //full screen
    if (document.documentElement.requestFullscreen) {
      document.documentElement.requestFullscreen();
    } else if (document.documentElement.mozRequestFullScreen) {
      document.documentElement.mozRequestFullScreen();
    } else if (document.documentElement.webkitRequestFullscreen) {
      document.documentElement.webkitRequestFullscreen(Element.ALLOW_KEYBOARD_INPUT);
    } else if (document.documentElement.msRequestFullscreen) {
      document.documentElement.msRequestFullscreen();
    }
  } else {
    //exit full screen
    if (document.exitFullscreen) {
      document.exitFullscreen();
    } else if (document.mozCancelFullScreen) {
      document.mozCancelFullScreen();
    } else if (document.webkitExitFullscreen) {
      document.webkitExitFullscreen();
    } else if (document.msExitFullscreen) {
      document.msExitFullscreen();
    }
  }
} // end ToggleFullScreen()

//***********************************************************************************************************************************************************************
//after the document page has loaded and is ready...
$(document).ready(function() {
  document.getElementById("vsnInfo").innerHTML ="Version: " + SW_VERSION;

  let urlParams = new URL(window.location.href).searchParams;

  // Override QLCP-NET server port if specified in the 'launch' URL
  let serverPortVal = parseInt(urlParams.get(SERVER_PORT_PARAM));
  if (! isNaN(serverPortVal)) {
    serverPort = serverPortVal;
  }

  // Override supported number of indication bits if specified in the 'launch' URL
  let numIndicationBitsVal = parseInt(urlParams.get(NUM_INDICATION_BITS_PARAM));
  if (! isNaN(numIndicationBitsVal)) {
    numIndicationBits = numIndicationBitsVal;
  }

  // Determine whether panel file request is to be 'automatically' initiated
  if (urlParams.get(AUTO_OPEN_PARAM)) {
    autoOpen = true
  }

  // Determine whether panel client is in 'read only' mode (disabling ability to edit/save)
  if (urlParams.get(READ_ONLY_PARAM)) {
    readOnly = true
  }

  InitGrid();

  InitDialogBoxes();

  // IIFE to periodically update the time.
  (function updateTime() {
    let d = new Date();
    let dateTimeStr = d.getFullYear() + '-' + monthNames[d.getMonth()] + '-' +
                      pad(d.getDate()) + '  ' + pad(d.getHours()) + ':' + pad(d.getMinutes());

    document.getElementById("system_time").innerHTML = dateTimeStr;

    setTimeout(updateTime, 10000);
  })();

  // If the browser supports web workers, create a web
  // worker to handle the QLCP-NET client communications.
  if (typeof(Worker) !== "undefined") {
    // Create the thread for QLCP-NET client communications.
    QnetClientThread = new Worker("scripts/qnet-client.js");

    // Add the event listener to process indication messages received from the QLCP-NET client comms thread.
    // Message data is received as msg.data object
    QnetClientThread.addEventListener("message", function(msg) {

      // If the link was down and has recovered.
      if (!linkOk && msg.data.link) {
        // Update the link status image
        document.getElementById("link_status").src = "images/linkOK24px.png";
        linkOk = true;
        SendAllControls();
      }

      // If the link was healthy and has failed.
      if (linkOk && !msg.data.link) {
        //update the link status image
        document.getElementById("link_status").src = "images/error24px.png";
        linkOk = false;
      }

      //Since this event will be fired periodically, it is a convenient time to
      //request a panel file via the developer tool if the user has clicked on the open file button.
      if (requestFileName) {
        $.get("http://localhost:"+serverPort+"/open/filename/", function(data){
          if (data !== "") {
            //success... open the panel file
            requestFileName = false;
            OpenPanel("http://localhost:"+serverPort+"/" + data);
          }
        }, "text");
      }
    });
  } else {
    // Hopefully this message will never be displayed when using the newer browsers.
    alert("The QLCP soft panel will not work with this browser because it does not support web workers.");
  }

  //Request the control labels (status names) from the QLCP-NET Developer tool and create the control status name options.
  let controlLabelsRequest = $.get("http://localhost:"+serverPort+"/control_labels/", function(data) {

    //Success! So enter design mode and prepare the UI accordingly.
    DesignMode();

    developerMode = true;

    //send a message to the web worker so that it knows the developer mode state and host url
    QnetClientThread.postMessage( {
      devMode:developerMode,
      url:hostUrl,
      serverPort:serverPort,
      sendControl:false,
      bitNbr:0,
      bitState:0,
      sendAllControls:false,
      cntrlBits:controlBits,
      numIndicationBits:numIndicationBits } );

    //update the control labels
    UpdateControlLabels(data);

    //Request the indication labels from the QLCP GUI Developer tool and create the indication status name options.
    $.get("http://localhost:"+serverPort+"/indication_labels/", function(data) {

        UpdateIndicationLabels(data);

    }, "text");

    //Request a panel file if 'auto open' was indicated at launch
    if (autoOpen) {
      RequestPanelFileDialog();
    }

  }, "text");

  //If there is no response from the developer tool, assume this is a live connection to the QLCP itself.
  //Request the location information and then the panel file from the QLCP server.
  controlLabelsRequest.fail(function() {

    // Ensure that components are created with positions locked.
    TogglePositionLock.positionLocked = true;

    // Request panel file and open
    OpenPanel("http://" + hostUrl + "/qlcp.cpf");

    //enter run mode
    RunMode();

    //hide the design mode button
    document.getElementById("design_button").style.visibility = "hidden";

    //send a message to the web worker so that it knows the developer mode state and host url
    QnetClientThread.postMessage( {
      devMode:developerMode,
      url:hostUrl,
      serverPort:serverPort,
      sendControl:false,
      bitNbr:0,
      bitState:0,
      sendAllControls:false,
      cntrlBits:controlBits,
      numIndicationBits:numIndicationBits } );

  });

  // This will periodically call UpdateComponentLayer to draw components as they are added/updated.
  window.requestAnimationFrame(UpdateComponentLayer);

  // Prevent context menu from popping up when user clicks on right mouse button
  document.addEventListener("contextmenu", function(e){
      e.preventDefault();
  }, false);
});

//This code supports the drag-n-drop feature for adding an lcp background image.
if(window.FileReader) {
  window.addEventListener('load', () => {
    let drop_zone = document.getElementById('component_container');

    // Tell the browser that we CAN drop on this target
    drop_zone.addEventListener('dragover', (e) => {
      e.preventDefault();
    }, false);
    drop_zone.addEventListener('drop', (e) => {
      e.preventDefault(); // stops the browser from redirecting off to the image.

      let files = e.dataTransfer.files;
      if (files.length > 0) {
        let file = files[0];
        if (typeof FileReader !== "undefined" && file.type.indexOf("image") != -1) {

          let reader = new FileReader();

          reader.onload = (e) => {

            // remove any background image for the panel that already exists
            bgImage_layer.removeChildren();
            bgImage_layer.clear();

            let bgImgObj = new Image();

            bgImgObj.onload = () => {

              let bgImg = new Konva.Image({
                image: bgImgObj,
                x: 20,
                y: 50,
                stroke: "rgba(0,255,255,1)",
                strokeWidth: 3,
                strokeEnabled: false,
                draggable: false,
                id: 'c' + ++idNbr // dkh tmp...not yet sure how image fits with other components
              });

              bgImage_layer.add(bgImg);
              bgImage_layer.draw();
            }
            bgImgObj.src = e.target.result;
          }

          reader.readAsDataURL(file);
        }
      }
    }, false);
  }, false);
} else {
  alert("Your browser does not support the HTML5 FileReader.");
}

//This function will delete or move selected component(s) using the keyboard.
//This function depends on requestAnimationFrame to redraw the component layer.
$(document).keydown(function (e) {

  if (designMode) {
    let allow_move = !TogglePositionLock.positionLocked;
    let step = e.shiftKey ? 10 : 1;

    let bgImg, bgImgActive = false;
    if (bgImage_layer.hasChildren()) {
      bgImg = bgImage_layer.getChildren()[0];
      if (bgImg.draggable()) {
        bgImgActive = true;
      }
    }

    //if this is just one selected component...
    if (selectedId !== "" && selectedComponents.length === 0) {
      let selObj = stage.find('#'+ selectedId)[0];
      let positionChanged = false;

      switch(e.key)
      {
        case "Delete":
          // Ignore if properties dialog is open
          if (!$( "#properties_dialog" ).dialog( "isOpen" )) {
            ClearPreviousSelection();
            QnetClientThread.removeEventListener("message", selObj.msgListener, false);
            selObj.destroy();
          }
          break;

        case "ArrowLeft":
          if (!$( "#properties_dialog" ).dialog( "isOpen" ) && allow_move) {
            e.preventDefault();
            selObj.position({
              x:  Math.max(selObj.attrs.x - step, 0),
              y:  selObj.attrs.y
            });
            positionChanged = true;
          }
          break;

        case "ArrowUp":
          if (!$( "#properties_dialog" ).dialog( "isOpen" ) && allow_move) {
            e.preventDefault();
            selObj.position({
              x:  selObj.attrs.x,
              y:  Math.max(selObj.attrs.y - step, 0)
            });
            positionChanged = true;
          }
          break;

        case "ArrowRight":
          if (!$( "#properties_dialog" ).dialog( "isOpen" ) && allow_move) {
            e.preventDefault();
            selObj.position({
              x:  Math.min(selObj.attrs.x + step, stage.attrs.width),
              y:  selObj.attrs.y
            });
            positionChanged = true;
          }
        break;

        case "ArrowDown":
          if (!$( "#properties_dialog" ).dialog( "isOpen" ) && allow_move) {
            e.preventDefault();
            selObj.position({
              x:  selObj.attrs.x,
              y:  Math.min(selObj.attrs.y + step, stage.attrs.height)
              });
            positionChanged = true;
          }
          break;

        case "Control":
          ctrlKeyDown = true;
          break;
      }

      ShowComponentPosition(selObj);

      if ($( "#properties_dialog" ).dialog( "isOpen" ) && positionChanged) {
        document.getElementById("x-pos-value").value = selObj.attrs.x;
        document.getElementById("y-pos-value").value = selObj.attrs.y;
      }
    } else if (bgImgActive) { // handle keystrokes on background image.
      switch (e.key) {
        case "Delete":
          e.preventDefault();
          bgImage_layer.removeChildren();
          bgImage_layer.clear();
          document.getElementById("position_display_textbox").value = "";
        break;

        case "ArrowLeft":
          e.preventDefault();
          bgImg.x(Math.max(bgImg.x() - step, 0));
          bgImage_layer.draw();
          document.getElementById("position_display_textbox").value = "  x = " +
          Math.round(bgImage_layer.getChildren()[0].x()) + ", y = " + Math.round(bgImage_layer.getChildren()[0].y());
        break;

        case "ArrowUp":
          e.preventDefault();
          bgImg.y(Math.max(bgImg.y() - step, 0));
          bgImage_layer.draw();
          document.getElementById("position_display_textbox").value = "  x = " +
          Math.round(bgImage_layer.getChildren()[0].x()) + ", y = " + Math.round(bgImage_layer.getChildren()[0].y());
        break;

        case "ArrowRight":
          e.preventDefault();
          bgImg.x(Math.min(bgImg.x() + step, stage.attrs.width));
          bgImage_layer.draw();
          document.getElementById("position_display_textbox").value = "  x = " +
          Math.round(bgImage_layer.getChildren()[0].x()) + ", y = " + Math.round(bgImage_layer.getChildren()[0].y());
        break;

        case "ArrowDown":
          e.preventDefault();
          bgImg.y(Math.min(bgImg.y() + step, stage.attrs.height));
          bgImage_layer.draw();
          document.getElementById("position_display_textbox").value = "  x = " +
          Math.round(bgImage_layer.getChildren()[0].x()) + ", y = " + Math.round(bgImage_layer.getChildren()[0].y());
        break;
      }
    } else if (selectedComponents.length !== 0) {
      //there is a selected group of components
      switch(e.key) {
        case "Delete":
          if ($( "#properties_dialog" ).dialog( "isOpen" )) {
            $( "#properties_dialog" ).dialog( "close" );
          } else {
            for (let i = 0; i < selectedComponents.length; i++) {
              QnetClientThread.removeEventListener("message", selectedComponents[i].msgListener, false);
              selectedComponents[i].destroy();
            }
            selectedComponents.length = 0;
            selectedId = "";
          }
          break;

        case "ArrowLeft":
          e.preventDefault();
          if (allow_move) {
            for (let i = 0; i < selectedComponents.length; i++) {
              selectedComponents[i].position({
                x: selectedComponents[i].getPosition().x - step,
                y: selectedComponents[i].getPosition().y
              });
            }
          }
          break;

        case "ArrowUp":     //move group up
          e.preventDefault();
          if (allow_move) {
            for (let i = 0; i < selectedComponents.length; i++) {
              selectedComponents[i].position({
                x: selectedComponents[i].getPosition().x,
                y: selectedComponents[i].getPosition().y - step
              });
            }
          }
          break;

        case "ArrowRight":     //move group right
          e.preventDefault();
          if (allow_move) {
            for (let i = 0; i < selectedComponents.length; i++) {
              selectedComponents[i].position({
                x: selectedComponents[i].getPosition().x + step,
                y: selectedComponents[i].getPosition().y
              });
            }
          }
          break;

        case "ArrowDown":     //move group down
          e.preventDefault();
          if (allow_move) {
            for (let i = 0; i < selectedComponents.length; i++) {
              selectedComponents[i].position({
                x: selectedComponents[i].getPosition().x,
                y: selectedComponents[i].getPosition().y + step
              });
            }
          }
          break;

        case "Control":
          ctrlKeyDown = true;
          break;
      }
    }
  }
});

$(document).keyup(function (e){
  switch (e.key) {
    case "Control":
      ctrlKeyDown = false;
    break;
  }
});

//Detect a change in full screen mode so that the button image and tool tip can be updated/toggled accordingly.
//This is required because F11 and the escape key can be used to change the mode, not just the button that calls ToggleFullScreen.
$(document).on('fullscreenchange webkitfullscreenchange mozfullscreenchange msfullscreenchange', function () {

  let fullscreenElement = document.fullscreenElement || document.webkitFullscreenElement || document.mozFullscreenElement || document.msFullscreenElement;

  if (!fullscreenElement) {
    //toggle button image and tool tip
    document.getElementById("fullscreen_button").title = "Full Screen";
    document.getElementById("fullscreen_image").src = "images/full_screen.png";
  } else {
    //toggle button image and tool tip
    document.getElementById("fullscreen_button").title = "Exit Full Screen";
    document.getElementById("fullscreen_image").src = "images/exit_full_screen.png";
  }
});

// dkh working on stage zoom (not browser zoom) ... so far, not good..
// let scl = 1;
// stage.on('wheel', (e) => {
//   if (e.evt.ctrlKey) {
//     e.evt.preventDefault();

//     let mult = e.evt.deltaY > 0 ? -1 : 1;
//     scl = scl + (0.01 * mult);
//     document.getElementById("scroll_container").style.transform = "scale(" + scl + ")";
//   }
// });